# Copyright 2017, Big Switch Networks
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
"""support multiple logical routers

Revision ID: 34f888285695
Revises: e6cb930d25de
Create Date: 2017-01-17 04:44:12.582865

"""

from alembic import op
from oslo_db.exception import DBError
import sqlalchemy as sa

# revision identifiers, used by Alembic.
revision = '34f888285695'
down_revision = '1ef57200f387'
branch_labels = None
depends_on = None


def upgrade():
    # add column tenant_id. allow nullable until we add values
    try:
        op.add_column('bsn_routerrules',
                      sa.Column('tenant_id', sa.String(255), nullable=True))
    except DBError as e:
        if ('Duplicate' not in e.message):
            raise e

    # populate it with the correct tenant_id from upstream table routers
    connection = op.get_bind()
    connection.execute("UPDATE bsn_routerrules bsn "
                       "INNER JOIN routers upstream "
                       "ON bsn.router_id = upstream.id "
                       "SET bsn.tenant_id = upstream.project_id;")

    # set the tenant_id column nullable to false
    op.alter_column('bsn_routerrules', 'tenant_id',
                    existing_type=sa.String(length=255), nullable=False)
    # update default rule priority. only to be executed if the constraint
    # addition succeeds in the previous statement.
    connection.execute("UPDATE bsn_routerrules "
                       "SET priority = 14000 "
                       "WHERE priority = 3000;")

    # drop the existing unique key constraint
    try:
        op.drop_constraint('unique_prio_rid', 'bsn_routerrules',
                           type_='unique')
    except Exception as e:
        pass

    # create unique key constraint named 'unique_prio_tid' on table
    # 'bsn_routerrules' including columns 'tenant_id' and 'priority'
    try:
        op.create_index('unique_prio_tid', 'bsn_routerrules',
                        ['tenant_id', 'priority'], unique=True)
    except Exception as e:
        pass


def downgrade():
    pass



# Copyright 2013 Big Switch Networks, Inc.
# All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os

import mock
import neutron.common.test_lib as test_lib
from oslo_config import cfg

from networking_bigswitch.plugins.bigswitch import config
from networking_bigswitch.plugins.bigswitch.db import consistency_db
from networking_bigswitch.plugins.bigswitch.db import network_template_db  # noqa
from networking_bigswitch.plugins.bigswitch.db import reachability_test_db  # noqa
from networking_bigswitch.plugins.bigswitch.db import tenant_policy_db  # noqa
from networking_bigswitch.tests.unit.bigswitch import fake_server
from networking_bigswitch.tests.unit.bigswitch.mock_paths import BACKGROUND
from networking_bigswitch.tests.unit.bigswitch.mock_paths import \
    BSN_SERVICE_PLUGIN_PATH
from networking_bigswitch.tests.unit.bigswitch.mock_paths import HTTPCON
from networking_bigswitch.tests.unit.bigswitch.mock_paths import \
    IS_UNICODE_ENABLED
from networking_bigswitch.tests.unit.bigswitch.mock_paths import \
    KEYSTONE_CLIENT
from networking_bigswitch.tests.unit.bigswitch.mock_paths import \
    LIB_RPC_TRANSPORT
from networking_bigswitch.tests.unit.bigswitch.mock_paths import \
    MAP_DISPLAY_NAME_OR_TENANT
from networking_bigswitch.tests.unit.bigswitch.mock_paths import PLUGIN_PATH
from networking_bigswitch.tests.unit.bigswitch.mock_paths import \
    RPC_SERVER_START
from networking_bigswitch.tests.unit.bigswitch.mock_paths import SERVER_MANAGER
from networking_bigswitch.tests.unit.bigswitch.mock_paths import SPAWN


class BigSwitchTestBase(object):

    _plugin_name = ('%s.NeutronRestProxyV2' % PLUGIN_PATH)
    _bsn_service_plugin_name = ('%s.BSNServicePlugin'
                                % BSN_SERVICE_PLUGIN_PATH)

    def setup_config_files(self):
        etc_path = os.path.join(os.path.dirname(__file__), 'etc')
        test_lib.test_config['config_files'] = [os.path.join(etc_path,
                                                'restproxy.ini.test')]
        self.addCleanup(cfg.CONF.reset)
        self.addCleanup(consistency_db.clear_db)
        config.register_config()
        # Only try SSL on SSL tests
        cfg.CONF.set_override('server_ssl', False, 'RESTPROXY')
        etc_path = os.path.join(os.path.dirname(__file__), 'etc')
        cfg.CONF.set_override('ssl_cert_directory',
                              os.path.join(etc_path, 'ssl'), 'RESTPROXY')
        # The mock interferes with HTTP(S) connection caching
        cfg.CONF.set_override('cache_connections', False, 'RESTPROXY')
        cfg.CONF.set_override('service_plugins', ['router'])
        cfg.CONF.set_override('add_meta_server_route', False, 'RESTPROXY')
        cfg.CONF.set_override('api_extensions_path', False)

    def map_tenant_name_side_effect(self, value):
        # for old tests, always map tenant name
        value['tenant_name'] = 'tenant_name'
        return value

    def is_unicode_enabled_side_effect(self):
        # for old tests, always return False
        return False

    def setup_patches(self):
        self.lib_rpc_transport_p = mock.patch(LIB_RPC_TRANSPORT)
        self.rpc_server_start_p = mock.patch(RPC_SERVER_START)
        # prevent any greenthreads from spawning
        self.spawn_p = mock.patch(SPAWN, new=lambda *args, **kwargs: None)
        # prevent the consistency watchdog and keystone sync from starting
        self.watch_p = mock.patch(BACKGROUND, new=lambda *args, **kwargs: None)
        # disable exception log to prevent json parse error from showing
        self.log_exc_p = mock.patch(SERVER_MANAGER + ".LOG.exception",
                                    new=lambda *args, **kwargs: None)

        self.ksclient_p = mock.patch(KEYSTONE_CLIENT)

        self.map_display_name_or_tenant_p = mock.patch(
            MAP_DISPLAY_NAME_OR_TENANT,
            side_effect=self.map_tenant_name_side_effect)
        self.is_unicode_enabled_p = mock.patch(
            IS_UNICODE_ENABLED,
            side_effect=self.is_unicode_enabled_side_effect)

        # start all mock patches
        self.log_exc_p.start()
        self.lib_rpc_transport_p.start()
        self.rpc_server_start_p.start()
        self.spawn_p.start()
        self.watch_p.start()
        self.ksclient_p.start()
        self.map_display_name_or_tenant_p.start()
        self.is_unicode_enabled_p.start()

    def startHttpPatch(self):
        self.httpPatch = mock.patch(HTTPCON,
                                    new=fake_server.HTTPConnectionMock)
        self.httpPatch.start()

    def setup_db(self):
        # setup the db engine and models for the consistency db
        consistency_db.setup_db()



# Copyright 2016 Big Switch Networks, Inc.  All rights reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#
# Adapted from neutron.tests.unit.db.test_agents_db.py

from neutron.tests.unit import testlib_api
from neutron_lib import context

from networking_bigswitch.plugins.bigswitch.db import network_template_db


class TestNetworkTemplateAssignmentDbMixin(testlib_api.SqlTestCase):
    def setUp(self):
        super(TestNetworkTemplateAssignmentDbMixin, self).setUp()
        self.context = context.get_admin_context()
        self.dbmixin = network_template_db.NetworkTemplateAssignmentDbMixin()
        self.template_dbmixin = network_template_db.NetworkTemplateDbMixin()

    def _assert_ref_fields_are_equal(self, reference, result):
        """Compare (key, value) pairs of a reference dict with the result

           Note: the result MAY have additional keys
        """

        for field, value in reference.items():
            self.assertEqual(value, result[field], field)

    def test_create_networktemplateassignment(self):
        # first create network template
        network_template_dict = {
            'networktemplate': {
                'name': 'template1',
                'body': 'template body 1'
            }
        }

        template_dbobj_dict = self.template_dbmixin.create_networktemplate(
            self.context, network_template_dict)

        # now create template assignment associated with the template
        network_template_assignment_dict = {
            'networktemplateassignment': {
                'tenant_id': 'admin_tenant_id',
                'template_id': template_dbobj_dict['id'],
                'stack_id': 'stack1'
            }
        }

        template = self.dbmixin.create_networktemplateassignment(
            self.context, network_template_assignment_dict)
        self._assert_ref_fields_are_equal(
            network_template_assignment_dict['networktemplateassignment'],
            template)

    def test_get_networktemplateassignment(self):
        pass

    def test_get_networktemplatesassignment(self):
        pass

    def test_update_networktemplateassignment(self):
        pass

    def test_delete_networktemplate(self):
        pass



# Copyright 2014 Big Switch Networks, Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""
This module manages configuration options
"""

from neutron.conf.agent import common as agconfig
from neutron_lib.api.definitions import portbindings
from neutron_lib.utils import net
from oslo_config import cfg

restproxy_opts = [
    cfg.ListOpt('servers', default=['localhost:8800'],
                help=_("A comma separated list of Arista CCF "
                       "servers and port numbers. The plugin proxies the "
                       "requests to the CCF server, "
                       "which performs the networking configuration. Only one"
                       "server is needed per deployment, but you may wish to"
                       "deploy multiple servers to support failover.")),
    cfg.StrOpt('server_auth', secret=True,
               help=_("The username and password for authenticating against "
                      " the CCF controller.")),
    cfg.BoolOpt('server_ssl', default=True,
                help=_("If True, Use SSL when connecting to the Big Switch or "
                       "Floodlight controller.")),
    cfg.BoolOpt('ssl_sticky', default=True,
                help=_("Trust and store the first certificate received for "
                       "each controller address and use it to validate future "
                       "connections to that address.")),
    cfg.BoolOpt('no_ssl_validation', default=False,
                help=_("Disables SSL certificate validation for controllers")),
    cfg.BoolOpt('cache_connections', default=True,
                help=_("Re-use HTTP/HTTPS connections to the controller.")),
    cfg.StrOpt('ssl_cert_directory',
               default='/etc/neutron/plugins/bigswitch/ssl',
               help=_("Directory containing ca_certs and host_certs "
                      "certificate directories.")),
    cfg.BoolOpt('sync_data', default=False,
                help=_("Sync data on connect")),
    cfg.BoolOpt('auto_sync_on_failure', default=True,
                help=_("If neutron fails to create a resource because "
                       "the backend controller doesn't know of a dependency, "
                       "the plugin automatically triggers a full data "
                       "synchronization to the controller.")),
    cfg.IntOpt('consistency_interval', default=60,
               help=_("Time between verifications that the backend controller "
                      "database is consistent with Neutron. (0 to disable)")),
    cfg.IntOpt('server_timeout', default=10,
               help=_("Maximum number of seconds to wait for proxy request "
                      "to connect and complete.")),
    cfg.IntOpt('thread_pool_size', default=4,
               help=_("Maximum number of threads to spawn to handle large "
                      "volumes of port creations.")),
    cfg.StrOpt('neutron_id', default='neutron-' + net.get_hostname(),
               deprecated_name='quantum_id',
               help=_("User defined identifier for this Neutron deployment")),
    cfg.BoolOpt('add_meta_server_route', default=True,
                help=_("Flag to decide if a route to the metadata server "
                       "should be injected into the VM")),
    cfg.StrOpt('auth_url', help=_("Authentication URL")),
    cfg.StrOpt('auth_user', help=_("Admin username")),
    cfg.StrOpt('auth_password', help=_("Admin password")),
    cfg.StrOpt('auth_tenant', help=_("Admin tenant name")),
    cfg.IntOpt('keystone_sync_interval', default=300,
               help=_("Time between that keystone queries to sync "
                      "Openstack tenants. (0 to disable)")),
    cfg.BoolOpt('sync_security_groups', default=False,
                help=_("Sync security group info to Big Cloud Fabric for "
                       "enhanced Testpath visibility.")),
    cfg.BoolOpt('naming_scheme_unicode', default=True,
                help=_("Configure whether or not to configure CCF "
                       "with unicode display-name.")),
    cfg.IntOpt('state_sync_interval', default=300,
               help=_("Configure how often the OSP plugin should update "
                      "cluster information on CCF"))
]
router_opts = [
    cfg.MultiStrOpt('tenant_default_router_rule', default=['*:any:any:permit'],
                    help=_("The default router rules installed in new tenant "
                           "routers. Repeat the config option for each rule. "
                           "Format is <tenant>:<source>:<destination>:<action>"
                           " Use an * to specify default for all tenants.")),
    cfg.IntOpt('max_router_rules', default=200,
               help=_("Maximum number of router rules")),
]
nova_opts = [
    cfg.StrOpt('vif_type', default='ovs',
               help=_("Virtual interface type to configure on "
                      "Nova compute nodes")),
]

VIF_TYPES = [
    portbindings.VIF_TYPE_UNBOUND,
    portbindings.VIF_TYPE_BINDING_FAILED,
    portbindings.VIF_TYPE_DISTRIBUTED,
    portbindings.VIF_TYPE_OVS,
    portbindings.VIF_TYPE_BRIDGE,
    portbindings.VIF_TYPE_OTHER,
    portbindings.VIF_TYPE_VHOST_USER,
    'iovisor', 'dvs', '802.1qbg', '802.1qbh', 'hyperv',
    'midonet', 'ib_hostdev', 'hw_web', 'vrouter',
]

PROVIDER_NETWORK_TYPE = 'provider:network_type'
PROVIDER_PHYSNET = 'provider:physical_network'

# SR-IOV related constants
UNSUPPORTED_VNIC_TYPES = [portbindings.VNIC_DIRECT_PHYSICAL]
VNIC_TYPE_SRIOV = [portbindings.VNIC_DIRECT]
SRIOV_ACTIVE_ACTIVE_MODE_PHYSNET_SUBSTR = 'BSN-ACTIVE-'
SRIOV_ACTIVE_PHYSNET = SRIOV_ACTIVE_ACTIVE_MODE_PHYSNET_SUBSTR + '1'

# Each VIF Type can have a list of nova host IDs that are fixed to that type
for i in VIF_TYPES:
    opt = cfg.ListOpt('node_override_vif_' + i, default=[],
                      help=_("Nova compute nodes to manually set VIF "
                             "type to %s") % i)
    nova_opts.append(opt)

# Add the vif types for reference later
nova_opts.append(cfg.ListOpt('vif_types',
                             default=VIF_TYPES,
                             help=_('List of allowed vif_type values.')))

agent_opts = [
    cfg.StrOpt('integration_bridge', default='br-int',
               help=_('Name of integration bridge on compute '
                      'nodes used for security group insertion.')),
    cfg.IntOpt('polling_interval', default=5,
               help=_('Seconds between agent checks for port changes')),
]


def register_config():
    cfg.CONF.register_opts(restproxy_opts, "RESTPROXY")
    cfg.CONF.register_opts(router_opts, "ROUTER")
    cfg.CONF.register_opts(nova_opts, "NOVA")
    cfg.CONF.register_opts(agent_opts, "RESTPROXYAGENT")
    # include for report_interval
    cfg.CONF.register_opts(agconfig.AGENT_STATE_OPTS, "AGENT")
    agconfig.register_root_helper(cfg.CONF)



# Copyright 2018 Big Switch Networks, Inc.
# All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""For storing paths used for mocking, this is used in various test files.
"""
# Keystone
KEYSTONE_CLIENT = 'keystoneclient.v3.client.Client'

# Neutron
NEUTRON_AGENT = 'neutron.agent'
OVS_BRIDGE = NEUTRON_AGENT + '.common.ovs_lib.OVSBridge'
PLUGIN_API = NEUTRON_AGENT + '.rpc.PluginApi'
CONSUMER_CREATE = NEUTRON_AGENT + '.rpc.create_consumers'

SG_RPC = NEUTRON_AGENT + '.securitygroups_rpc'

CONTEXT = 'neutron_lib.context'

NEUTRON_CFG = 'neutron.common.config'
LIB_RPC_TRANSPORT = 'neutron_lib.rpc.TRANSPORT'

# oslo_messaging
RPC_SERVER_START = ('oslo_messaging.server.MessageHandlingServer.start')

# BSN
BSN_DIR = 'networking_bigswitch.plugins.bigswitch'

# Config
PL_CONFIG = BSN_DIR + '.config'

# DB
CONSISTENCY_DB = BSN_DIR + '.db.consistency_db'

# Driver
DRIVER_MOD = 'networking_bigswitch.plugins.ml2.drivers.mech_bigswitch.driver'
DRIVER = DRIVER_MOD + '.BigSwitchMechanismDriver'


# plugin path
PLUGIN_PATH = BSN_DIR + '.plugin'
BSN_SERVICE_PLUGIN_PATH = BSN_DIR + '.bsn_service_plugin'

NOTIFIER = PLUGIN_PATH + '.AgentNotifierApi'
SPAWN = PLUGIN_PATH + '.eventlet.GreenPool.spawn_n'
MAP_DISPLAY_NAME_OR_TENANT = (PLUGIN_PATH + '.NeutronRestProxyV2Base'
                              '._map_display_name_or_tenant')

# Agent
AGENT_MOD = BSN_DIR + '.agent.restproxy_agent'
SG_AGENT = AGENT_MOD + '.FilterDeviceIDMixin'

# SERVER MANAGER
SERVER_MANAGER = BSN_DIR + '.servermanager'
SERVER_REST_CALL = SERVER_MANAGER + '.ServerProxy.rest_call'
HTTPCON = SERVER_MANAGER + '.http_client.HTTPConnection'
HTTPSCON = SERVER_MANAGER + '.http_client.HTTPSConnection'

SERVER_POOL = SERVER_MANAGER + '.ServerPool'
POOL_REST_ACTION = SERVER_POOL + '.rest_action'
POOL_REST_CALL = SERVER_POOL + '.rest_call'
BACKGROUND = SERVER_POOL + '.start_background_tasks'
POOL_TOPO_SYNC = SERVER_POOL + '.force_topo_sync'
POOL_UPDATE_TENANT_CACHE = SERVER_POOL + '._update_tenant_cache'
POOL_GET_CAPABILITIES = SERVER_POOL + '.get_capabilities'
IS_UNICODE_ENABLED = SERVER_POOL + '.is_unicode_enabled'

# SSL Cert Related
GET_SERVER_CERTIFICATE = SERVER_MANAGER + '.ssl.get_server_certificate'
SSL_CREATE_DEFAULT_CONTEXT = SERVER_MANAGER + '.ssl.create_default_context'
CERT_COMBINER = SERVER_POOL + '._combine_certs_to_file'
FILE_PUT = SERVER_POOL + '._file_put_contents'
GET_CA_CERTS = SERVER_POOL + '._get_ca_cert_paths'



# Copyright 2011 OpenStack Foundation.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from neutron.api import extensions as neutron_extensions
from neutron.api.v2 import base
from neutron.api.v2 import resource_helper
from neutron_lib.api import extensions
from neutron_lib.plugins import directory

import abc
from networking_bigswitch.plugins.bigswitch import constants
from networking_bigswitch.plugins.bigswitch import extensions as bsn_extensions

# Ensure the extension is loaded at startup
neutron_extensions.append_api_extensions_path(bsn_extensions.__path__)

RESOURCE_ATTRIBUTE_MAP = {
    'networktemplates': {
        'id': {'allow_post': False, 'allow_put': False,
               'validate': {'type:string': None},
               'is_visible': True},
        'tenant_id': {'allow_post': True, 'allow_put': False,
                      'validate': {'type:string': None},
                      'is_visible': False},
        'body': {'allow_post': True, 'allow_put': True,
                 'validate': {'type:string': None},
                 'is_visible': True},
        'name': {'allow_post': True, 'allow_put': True,
                 'validate': {'type:string': None},
                 'is_visible': True}
    },
    'networktemplateassignments': {
        'id': {'allow_post': False, 'allow_put': False,
               'validate': {'type:string': None},
               'is_visible': True},
        'template_id': {'allow_post': True, 'allow_put': True,
                        'validate': {'type:string': None},
                        'is_visible': True},
        'tenant_id': {'allow_post': True, 'allow_put': False,
                      'validate': {'type:string': None},
                      'is_visible': True},
        'stack_id': {'allow_post': True, 'allow_put': True,
                     'validate': {'type:string': None},
                     'is_visible': True},
        'template': {'allow_post': False, 'allow_put': False,
                     'is_visible': True},
    },
    'reachabilitytests': {
        'id': {'allow_post': False, 'allow_put': False,
               'is_visible': True},
        'tenant_id': {'allow_post': True, 'allow_put': False,
                      'validate': {'type:string': None},
                      'is_visible': True},
        'name': {'allow_post': True, 'allow_put': True,
                 'validate': {'type:string': None},
                 'is_visible': True},
        'src_tenant_id': {'allow_post': True, 'allow_put': True,
                          'validate': {'type:string': None},
                          'is_visible': True},
        'src_tenant_name': {'allow_post': True, 'allow_put': True,
                            'validate': {'type:string': None},
                            'is_visible': True},
        'src_segment_id': {'allow_post': True, 'allow_put': True,
                           'validate': {'type:string': None},
                           'is_visible': True},
        'src_segment_name': {'allow_post': True, 'allow_put': True,
                             'validate': {'type:string': None},
                             'is_visible': True},
        'src_ip': {'allow_post': True, 'allow_put': True,
                   'is_visible': True},
        'dst_ip': {'allow_post': True, 'allow_put': True,
                   'is_visible': True},
        'expected_result': {'allow_post': True, 'allow_put': True,
                            'validate': {'type:string': None},
                            'is_visible': True},
        'test_time': {'allow_post': False, 'allow_put': False,
                      'is_visible': True},
        'test_result': {'allow_post': False, 'allow_put': False,
                        'is_visible': True},
        'detail': {'allow_post': False, 'allow_put': False,
                   'is_visible': True},
        'logical_path': {'allow_post': False, 'allow_put': False,
                         'is_visible': True},
        'run_test': {'allow_post': False, 'allow_put': True,
                     'is_visible': True},
    },
    'reachabilityquicktests': {
        'id': {'allow_post': False, 'allow_put': False,
               'is_visible': True},
        'tenant_id': {'allow_post': True, 'allow_put': False,
                      'validate': {'type:string': None},
                      'is_visible': True},
        'name': {'allow_post': True, 'allow_put': True,
                 'validate': {'type:string': None},
                 'is_visible': True},
        'src_tenant_id': {'allow_post': True, 'allow_put': True,
                          'validate': {'type:string': None},
                          'is_visible': True},
        'src_tenant_name': {'allow_post': True, 'allow_put': True,
                            'validate': {'type:string': None},
                            'is_visible': True},
        'src_segment_id': {'allow_post': True, 'allow_put': True,
                           'validate': {'type:string': None},
                           'is_visible': True},
        'src_segment_name': {'allow_post': True, 'allow_put': True,
                             'validate': {'type:string': None},
                             'is_visible': True},
        'src_ip': {'allow_post': True, 'allow_put': True,
                   'is_visible': True},
        'dst_ip': {'allow_post': True, 'allow_put': True,
                   'is_visible': True},
        'expected_result': {'allow_post': True, 'allow_put': True,
                            'validate': {'type:string': None},
                            'is_visible': True},
        'test_time': {'allow_post': False, 'allow_put': False,
                      'is_visible': True},
        'test_result': {'allow_post': False, 'allow_put': False,
                        'is_visible': True},
        'detail': {'allow_post': False, 'allow_put': False,
                   'is_visible': True},
        'logical_path': {'allow_post': False, 'allow_put': False,
                         'is_visible': True},
        'run_test': {'allow_post': False, 'allow_put': True,
                     'is_visible': True},
        'save_test': {'allow_post': False, 'allow_put': True,
                      'is_visible': True},
    },
    'tenantpolicies': {
        'id': {'allow_post': False, 'allow_put': False,
               'is_visible': True},
        'tenant_id': {'allow_post': True, 'allow_put': False,
                      'validate': {'type:string': None},
                      'is_visible': True},
        'priority': {'allow_post': True, 'allow_put': False,
                     'validate': {'type:integer': None},
                     'is_visible': True},
        'source': {'allow_post': True, 'allow_put': True,
                   'validate': {'type:string': None},
                   'is_visible': True},
        'source_port': {'allow_post': True, 'allow_put': True,
                        'validate': {'type:integer': None},
                        'is_visible': True,
                        'default': 0},
        'destination': {'allow_post': True, 'allow_put': True,
                        'validate': {'type:string': None},
                        'is_visible': True},
        'destination_port': {'allow_post': True, 'allow_put': True,
                             'validate': {'type:integer': None},
                             'is_visible': True, 'default': 0},
        'protocol': {'allow_post': True, 'allow_put': True,
                     'validate': {'type:string': None},
                     'is_visible': True, 'default': ''},
        'action': {'allow_post': True, 'allow_put': True,
                   'validate': {'type:string': None},
                   'is_visible': True},
        'nexthops': {'allow_post': True, 'allow_put': True,
                     'validate': {'type:string': None},
                     'is_visible': True,
                     'default': ''},
    },
    'forcesynctopologies': {
        'id': {'allow_post': False, 'allow_put': False,
               'is_visible': True},
        'project_id': {'allow_post': True, 'allow_put': False,
                       'validate': {'type:string': None},
                       'is_visible': True},
        'timestamp_ms': {'allow_post': False, 'allow_put': True,
                         'validate': {'type:string': None},
                         'is_visible': True},
        'timestamp_datetime': {'allow_post': False, 'allow_put': False,
                               'validate': {'type:string': None},
                               'is_visible': True},
        'status': {'allow_post': False, 'allow_put': False,
                   'validate': {'type:string': None},
                   'is_visible': True}
    },
}


class Bsnserviceextension(extensions.ExtensionDescriptor):

    """API extension for handling BSN service plugin."""

    @classmethod
    def get_name(cls):
        return "Big Switch Networks Service Plugin"

    @classmethod
    def get_alias(cls):
        return "bsn-service-extension"

    @classmethod
    def get_description(cls):
        return "Provides REST API for BSN specific service extensions."

    @classmethod
    def get_namespace(cls):
        return \
            "http://docs.openstack.org/ext/neutron/bsnserviceplugin/api/v1.0"

    @classmethod
    def get_updated(cls):
        return "2015-10-29T04:20:00-00:00"

    @classmethod
    def get_resources(cls):
        """Returns Extended Resources."""
        resources = []
        net_template_inst = directory.get_plugin(constants.BSN_SERVICE_PLUGIN)
        plural_mappings = resource_helper.build_plural_mappings(
            {}, RESOURCE_ATTRIBUTE_MAP)
        for collection in RESOURCE_ATTRIBUTE_MAP:
            controller = base.create_resource(
                collection, plural_mappings[collection], net_template_inst,
                RESOURCE_ATTRIBUTE_MAP[collection])
            resource = neutron_extensions.ResourceExtension(
                collection, controller)
            resources.append(resource)
        return resources

    def get_extended_resources(self, version):
        if version == "2.0":
            return RESOURCE_ATTRIBUTE_MAP
        else:
            return {}


class BSNServicePluginBase(object):

    # Network Template
    @abc.abstractmethod
    def get_networktemplates(self, context, filters=None, fields=None,
                             sorts=None, limit=None, marker=None,
                             page_reverse=False):
        pass

    @abc.abstractmethod
    def get_networktemplate(self, context, id, fields=None):
        pass

    @abc.abstractmethod
    def create_networktemplate(self, context, networktemplate):
        pass

    @abc.abstractmethod
    def delete_networktemplate(self, context, id):
        pass

    @abc.abstractmethod
    def update_networktemplate(self, context, id, networktemplate):
        pass

    # Network Template Assignment
    @abc.abstractmethod
    def get_networktemplateassignments(self, context, filters=None,
                                       fields=None, sorts=None, limit=None,
                                       marker=None, page_reverse=False):
        pass

    @abc.abstractmethod
    def get_networktemplateassignment(self, context, id, fields=None):
        pass

    @abc.abstractmethod
    def create_networktemplateassignment(self, context,
                                         networktemplateassignment):
        pass

    @abc.abstractmethod
    def delete_networktemplateassignment(self, context, id):
        pass

    @abc.abstractmethod
    def update_networktemplateassignment(self, context, id,
                                         networktemplateassignment):
        pass

    # Reachability Test
    @abc.abstractmethod
    def get_reachabilitytests(self, context, filters=None, fields=None,
                              sorts=None, limit=None, marker=None,
                              page_reverse=False):
        pass

    @abc.abstractmethod
    def get_reachabilitytest(self, context, id, fields=None):
        pass

    @abc.abstractmethod
    def create_reachabilitytest(self, context, reachabilitytest):
        pass

    @abc.abstractmethod
    def update_reachabilitytest(self, context, id, reachabilitytest):
        pass

    @abc.abstractmethod
    def delete_reachabilitytest(self, context, id):
        pass

    # Reachability Quick Test
    @abc.abstractmethod
    def get_reachabilityquicktests(self, context, filters=None, fields=None,
                                   sorts=None, limit=None, marker=None,
                                   page_reverse=False):
        pass

    @abc.abstractmethod
    def get_reachabilityquicktest(self, context, id, fields=None):
        pass

    @abc.abstractmethod
    def create_reachabilityquicktest(self, context, reachabilityquicktest):
        pass

    @abc.abstractmethod
    def update_reachabilityquicktest(self, context, id, reachabilityquicktest):
        pass

    @abc.abstractmethod
    def delete_reachabilityquicktest(self, context, id):
        pass

    # Tenant router policies
    @abc.abstractmethod
    def get_tenantpolicies(self, context, filters=None, fields=None,
                           sorts=None, limit=None, marker=None,
                           page_reverse=False):
        pass

    @abc.abstractmethod
    def get_tenantpolicy(self, context, id, fields=None):
        pass

    @abc.abstractmethod
    def create_tenantpolicy(self, context, tenantpolicy):
        pass

    @abc.abstractmethod
    def update_tenantpolicy(self, context, id, tenantpolicy):
        pass

    @abc.abstractmethod
    def delete_tenantpolicy(self, context, id):
        pass

    # Force Topology Sync command
    @abc.abstractmethod
    def update_forcesynctopology(self, context, id, forcesynctopology):
        pass

    @abc.abstractmethod
    def get_forcesynctopologies(self, context, filters=None, fields=None,
                                sorts=None, limit=None, marker=None,
                                page_reverse=False):
        pass

    @abc.abstractmethod
    def get_forcesynctopology(self, context, id, fields=None):
        pass



#!/usr/bin/python
# Copyright 2016 Big Switch Networks, Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#

from distutils.version import StrictVersion
import sys

# read the two git diff lines about version
two_lines = sys.stdin.read()
if 'version' not in two_lines:
    sys.exit("version not found in args. Build FAILED")
lines = str(two_lines).split('\n')
version1 = StrictVersion(lines[0].split('=')[1].strip())
version2 = StrictVersion(lines[1].split('=')[1].strip())
print('version1: ', version1)
print('version2: ', version2)
if version2 > version1:
    print('Version update correct.')
else:
    sys.exit("new version string is < old version string. Build FAILED")



# Copyright 2014 Big Switch Networks, Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import mock
from neutron.tests import base
from oslo_utils import importutils

from networking_bigswitch.plugins.bigswitch import config as pl_config
from networking_bigswitch.tests.unit.bigswitch.mock_paths import AGENT_MOD
from networking_bigswitch.tests.unit.bigswitch.mock_paths import \
    CONSUMER_CREATE
from networking_bigswitch.tests.unit.bigswitch.mock_paths import CONTEXT
from networking_bigswitch.tests.unit.bigswitch.mock_paths import NEUTRON_CFG
from networking_bigswitch.tests.unit.bigswitch.mock_paths import OVS_BRIDGE
from networking_bigswitch.tests.unit.bigswitch.mock_paths import PL_CONFIG
from networking_bigswitch.tests.unit.bigswitch.mock_paths import PLUGIN_API
from networking_bigswitch.tests.unit.bigswitch.mock_paths import SG_AGENT
from networking_bigswitch.tests.unit.bigswitch.mock_paths import SG_RPC


class BaseAgentTestCase(base.BaseTestCase):

    def setUp(self):
        pl_config.register_config()
        super(BaseAgentTestCase, self).setUp()
        self.mod_agent = importutils.import_module(AGENT_MOD)


class TestRestProxyAgentOVS(BaseAgentTestCase):
    def setUp(self):
        super(TestRestProxyAgentOVS, self).setUp()
        self.plapi = mock.patch(PLUGIN_API).start()
        self.ovsbridge_p = mock.patch(OVS_BRIDGE)
        self.ovsbridge = self.ovsbridge_p.start()
        self.context = mock.patch(CONTEXT).start()
        self.rpc = mock.patch(CONSUMER_CREATE).start()
        self.sg_agent = mock.patch(SG_AGENT).start()
        self.sg_rpc = mock.patch(SG_RPC).start()

    def mock_agent(self):
        mock_context = mock.Mock(return_value='abc')
        self.context.get_admin_context_without_session = mock_context
        return self.mod_agent.RestProxyAgent('int-br', 2)

    def mock_port_update(self, **kwargs):
        agent = self.mock_agent()
        agent.port_update(mock.Mock(), **kwargs)

    def test_port_update(self):
        port = {'id': '1', 'security_groups': 'default'}

        with mock.patch.object(self.ovsbridge.return_value,
                               'get_vif_port_by_id',
                               return_value='1') as get_vif:
            self.mock_port_update(port=port)

        get_vif.assert_called_once_with('1')
        self.sg_agent.assert_has_calls([
            mock.call().refresh_firewall()
        ])

    def test_port_update_not_vifport(self):
        port = {'id': '1', 'security_groups': 'default'}

        with mock.patch.object(self.ovsbridge.return_value,
                               'get_vif_port_by_id',
                               return_value=False) as get_vif:
            self.mock_port_update(port=port)

        get_vif.assert_called_once_with('1')
        self.assertFalse(self.sg_agent.return_value.refresh_firewall.called)

    def test_port_update_without_secgroup(self):
        port = {'id': '1'}

        with mock.patch.object(self.ovsbridge.return_value,
                               'get_vif_port_by_id',
                               return_value='1') as get_vif:
            self.mock_port_update(port=port)

        get_vif.assert_called_once_with('1')
        self.assertFalse(self.sg_agent.return_value.refresh_firewall.called)

    def mock_update_ports(self, vif_port_set=None, registered_ports=None):
        vif_port_set = vif_port_set or set()
        registered_ports = registered_ports or set()
        with mock.patch.object(self.ovsbridge.return_value,
                               'get_vif_port_set',
                               return_value=vif_port_set):
            agent = self.mock_agent()
            return agent._update_ports(registered_ports)

    def test_update_ports_unchanged(self):
        self.assertIsNone(self.mock_update_ports())

    def test_update_ports_changed(self):
        vif_port_set = set(['1', '3'])
        registered_ports = set(['1', '2'])
        expected = dict(current=vif_port_set,
                        added=set(['3']),
                        removed=set(['2']))

        actual = self.mock_update_ports(vif_port_set, registered_ports)

        self.assertEqual(expected, actual)

    def mock_process_devices_filter(self, port_info):
        agent = self.mock_agent()
        agent._process_devices_filter(port_info)

    def test_process_devices_filter_add(self):
        port_info = {'added': 1}

        self.mock_process_devices_filter(port_info)

        self.sg_agent.assert_has_calls([
            mock.call().prepare_devices_filter(1)
        ])

    def test_process_devices_filter_remove(self):
        port_info = {'removed': 2}

        self.mock_process_devices_filter(port_info)

        self.sg_agent.assert_has_calls([
            mock.call().remove_devices_filter(2)
        ])

    def test_process_devices_filter_both(self):
        port_info = {'added': 1, 'removed': 2}

        self.mock_process_devices_filter(port_info)

        self.sg_agent.assert_has_calls([
            mock.call().prepare_devices_filter(1),
            mock.call().remove_devices_filter(2)
        ])

    def test_process_devices_filter_none(self):
        port_info = {}

        self.mock_process_devices_filter(port_info)

        self.assertFalse(
            self.sg_agent.return_value.prepare_devices_filter.called)
        self.assertFalse(
            self.sg_agent.return_value.remove_devices_filter.called)


class TestRestProxyAgent(BaseAgentTestCase):
    def mock_main(self):
        cfg_attrs = {'CONF.RESTPROXYAGENT.integration_bridge': 'integ_br',
                     'CONF.RESTPROXYAGENT.polling_interval': 5,
                     'CONF.AGENT.root_helper': 'helper',
                     'CONF.AGENT.report_interval': 60}
        with\
            mock.patch(AGENT_MOD + '.cfg', **cfg_attrs) as mock_conf,\
            mock.patch(AGENT_MOD + '.config.init'),\
            mock.patch(NEUTRON_CFG) as mock_log_conf,\
                mock.patch(PL_CONFIG):
            self.mod_agent.main()

        mock_log_conf.assert_has_calls([
            mock.call(mock_conf),
        ])

    def test_main(self):
        agent_attrs = {'daemon_loop.side_effect': SystemExit(0)}
        with mock.patch(AGENT_MOD + '.RestProxyAgent',
                        **agent_attrs) as mock_agent:
            self.assertRaises(SystemExit, self.mock_main)

        mock_agent.assert_has_calls([
            mock.call('integ_br', 5),
            mock.call().daemon_loop()
        ])



# Copyright 2017, Big Switch Networks
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import sqlalchemy as sa

from networking_bigswitch.plugins.bigswitch.i18n import _
from networking_bigswitch.plugins.bigswitch.utils import Util
from neutron.db.models import l3 as l3_models
from neutron_lib.api import validators
from neutron_lib.db import api as db_api
from neutron_lib.db import model_base
from neutron_lib.db import model_query
from neutron_lib.db import utils as db_utils
from neutron_lib import exceptions as n_exc
from oslo_config import cfg
from oslo_db import exception as db_exc
from oslo_log import log as logging
from oslo_utils import uuidutils
from sqlalchemy.orm import exc, relationship
from sqlalchemy.types import Enum

LOG = logging.getLogger(__name__)

DEFAULT_POLICY_PRIORITY = 14000
# number of fields in a router rule string
ROUTER_RULE_COMPONENT_COUNT = 5


class TenantPolicyException(n_exc.NeutronException):
    message = _("Error in tenant policy operation: %(error_msg)s")
    status = None

    def __init__(self, **kwargs):
        self.tenant = kwargs.get('error_msg')
        super(TenantPolicyException, self).__init__(**kwargs)


class TenantPolicy(model_base.BASEV2,
                   model_base.HasId,
                   model_base.HasProject):
    __tablename__ = 'bsn_tenantpolicies'
    priority = sa.Column(sa.Integer, nullable=False)
    source = sa.Column(sa.String(255), nullable=False)
    source_port = sa.Column(sa.Integer, nullable=True)
    destination = sa.Column(sa.String(255), nullable=False)
    destination_port = sa.Column(sa.Integer, nullable=True)
    protocol = sa.Column(Enum("tcp", "udp"), nullable=True)
    action = sa.Column(Enum("deny", "permit", name="action"), nullable=False)
    nexthops = relationship('TenantPolicyNextHop',
                            cascade='all,delete,delete-orphan')

    class Meta(object):
        unique_together = ('priority', 'tenant_id')


class TenantPolicyNextHop(model_base.BASEV2):
    __tablename__ = 'bsn_tenantpolicy_nexthops'
    policy_id = sa.Column(sa.String(36), sa.ForeignKey('bsn_tenantpolicies.id',
                                                       ondelete="CASCADE"),
                          primary_key=True)
    nexthop = sa.Column(sa.String(255), nullable=False, primary_key=True)


class TenantPolicyNotFound(n_exc.NotFound):
    message = _("Tenant Policy %(id)s could not be found")


class TenantPolicyExists(n_exc.NeutronException):
    message = _("Tenant policy for tenant ID %(tenant_id)s with "
                "priority %(priority)s exists")


class TenantRouterDoesNotExist(n_exc.NotFound):
    message = _("Tenant does not have a router to apply policy. "
                "Please create a router before adding policies.")


class TenantPolicyDbMixin(object):
    # internal methods
    def _make_tenantpolicy_dict(self, tenantpolicy, fields=None):
        nexthops = [hop['nexthop'] for hop in tenantpolicy.nexthops]
        return db_utils.resource_fields({
            'id': tenantpolicy.id,
            'tenant_id': tenantpolicy.tenant_id,
            'priority': tenantpolicy.priority,
            'source': tenantpolicy.source,
            'source_port': tenantpolicy.source_port,
            'destination': tenantpolicy.destination,
            'destination_port': tenantpolicy.destination_port,
            'protocol': tenantpolicy.protocol,
            'action': tenantpolicy.action,
            'nexthops': nexthops}, fields)

    def _get_tenantpolicy(self, context, id):
        try:
            tenantpolicy = model_query.get_by_id(context, TenantPolicy, id)
        except exc.NoResultFound:
            raise TenantPolicyNotFound(id=id)
        return tenantpolicy

    def _validate_port_number(self, port_num):
        msg = None
        int_port_num = int(port_num)
        if int_port_num not in range(0, 65536):
            msg = (_("Port number specified %s in policy not in valid "
                     "range of 1 to 65535.") % port_num)
        return msg

    def _validate_nexthops(self, nexthops):
        msg = None
        seen = []
        nexthops = nexthops.split(',')
        for ip in nexthops:
            if not ip:
                continue
            msg = validators.validate_ip_address(ip)
            if ip in seen:
                msg = _("Duplicate nexthop in rule '%s'") % ip
                return msg
            seen.append(ip)
        return msg

    def _validate_action(self, action):
        msg = None
        if action not in ['permit', 'deny']:
            msg = _("Action must be either permit or deny."
                    " '%s' was provided") % action
        return msg

    def _validate_priority(self, priority):
        msg = None
        int_priority = int(priority)
        if (int_priority not in range(1, 3001) and
                int_priority != DEFAULT_POLICY_PRIORITY):
            msg = _("User must provide valid priority between 1 and 3000. "
                    "%s was provided.") % priority
        return msg

    def _validate_port_protocol(self, policy_data):
        msg = None
        if ((int(policy_data['source_port']) > 0 or
                int(policy_data['destination_port']) > 0) and
                policy_data['protocol'] not in ['tcp', 'udp']):
            msg = _("Protocol must be select if either source or destination "
                    "port is specified.")
        return msg

    def _cleanse_policy(self, policy_data):
        """Some values are given defaults via the REST API,

        but those should be set to NULL when going to the DB.

        This method normalizes those values.

        """
        if int(policy_data['source_port']) == 0:
            policy_data['source_port'] = None
        if int(policy_data['destination_port']) == 0:
            policy_data['destination_port'] = None
        if policy_data['protocol'] == '':
            policy_data['protocol'] = None
        if policy_data['nexthops']:
            policy_data['nexthops'] = policy_data['nexthops'].split(',')
        return policy_data

    def _validate_and_cleanse_policy(self, context, policy_data):
        """Validate every field of the policy and raise exceptions if any.

        This is a simple syntax validation. Actual funtional validation is
        performed at the backend by BCF and error message returned is bubbled
        up to the Horizon GUI.

        :param context: context of the transaction
        :param policy_data: the policy resource to be validated
        """
        policy_data['tenant_id'] = Util.get_tenant_id_for_create(
            context, policy_data)
        V4ANY = '0.0.0.0/0'
        CIDRALL = ['any', 'external']
        source = (V4ANY if policy_data['source'] in CIDRALL
                  else policy_data['source'])
        destination = (V4ANY if policy_data['destination'] in CIDRALL
                       else policy_data['destination'])
        errors = [validators.validate_subnet(source),
                  validators.validate_subnet(destination),
                  self._validate_nexthops(policy_data['nexthops']),
                  self._validate_action(policy_data['action']),
                  self._validate_priority(policy_data['priority']),
                  self._validate_port_number(policy_data['source_port']),
                  self._validate_port_number(policy_data['destination_port']),
                  self._validate_port_protocol(policy_data)]
        errors = [m for m in errors if m]
        if errors:
            LOG.debug(errors)
            raise n_exc.InvalidInput(error_message=errors)

        return self._cleanse_policy(policy_data)

    # public CRUD methods for network templates
    def get_tenantpolicies(self, context, filters=None, fields=None,
                           sorts=None, limit=None, marker=None,
                           page_reverse=False):
        with db_api.CONTEXT_READER.using(context):
            tenantpolicies = \
                model_query.get_collection(context, TenantPolicy,
                                           self._make_tenantpolicy_dict,
                                           filters=filters, fields=fields)
        return tenantpolicies

    def get_tenantpolicy(self, context, id, fields=None):
        with db_api.CONTEXT_READER.using(context):
            tenantpolicy = self._get_tenantpolicy(context, id)
            return self._make_tenantpolicy_dict(tenantpolicy, fields)

    @db_api._tag_retriables_as_unretriable
    def create_tenantpolicy(self, context, tenantpolicy):
        tenantpolicy_data = tenantpolicy['tenantpolicy']
        tenantpolicy_data = self._validate_and_cleanse_policy(
            context, tenantpolicy_data)
        with db_api.CONTEXT_WRITER.using(context):
            router_exists = context.session.query(l3_models.Router).filter_by(
                project_id=tenantpolicy_data['project_id']).first()
            if not router_exists:
                raise TenantRouterDoesNotExist()

            tenantpolicy = TenantPolicy(
                id=uuidutils.generate_uuid(),
                tenant_id=tenantpolicy_data['tenant_id'],
                priority=tenantpolicy_data['priority'],
                source=tenantpolicy_data['source'],
                source_port=tenantpolicy_data['source_port'],
                destination=tenantpolicy_data['destination'],
                destination_port=tenantpolicy_data['destination_port'],
                protocol=tenantpolicy_data['protocol'],
                action=tenantpolicy_data['action'],
                nexthops=[TenantPolicyNextHop(nexthop=hop)
                          for hop in tenantpolicy_data['nexthops']])
            try:
                context.session.add(tenantpolicy)
            except db_exc.DBDuplicateEntry:
                raise TenantPolicyExists(tenant_id=tenantpolicy.project_id,
                                         priority=tenantpolicy.priority)
        return self._make_tenantpolicy_dict(tenantpolicy)

    def delete_tenantpolicy(self, context, id):
        with db_api.CONTEXT_WRITER.using(context):
            tenantpolicy = self._get_tenantpolicy(context, id)
            context.session.delete(tenantpolicy)

    def update_tenantpolicy(self, context, servers, id, tenantpolicy):
        tenantpolicy_data = tenantpolicy['tenantpolicy']
        tenantpolicy_data = self._validate_and_cleanse_policy(
            context, tenantpolicy_data)
        with db_api.CONTEXT_WRITER.using(context):
            tenantpolicy = self._get_tenantpolicy(context, id)
            tenantpolicy.update(tenantpolicy_data)
        return self._make_tenantpolicy_dict(tenantpolicy)

    # handle default policy during create and delete of routers
    def _get_tenant_default_router_policy(self, tenant):
        """Returns a rule dictionary. Can be empty.

        """
        rules = cfg.CONF.ROUTER.tenant_default_router_rule
        default_rule = {}
        tenant_rule = {}
        for rule in rules:
            items = rule.split(':')
            # put an empty string on the end if nexthops wasn't specified
            if len(items) < ROUTER_RULE_COMPONENT_COUNT:
                items.append('')
            try:
                (tenant_id, source, destination, action, nexthops) = items
            except ValueError:
                continue
            parsed_rule = {'priority': -1,
                           'source': source,
                           'destination': destination,
                           'action': action,
                           'nexthops': [hop for hop in nexthops.split(',')
                                        if hop]}
            if tenant_id == '*':
                if default_rule:
                    # a default rule already exists
                    raise TenantPolicyException(
                        error_msg='Number of default rules exceeds the '
                                  'limit of 1!')
                default_rule = parsed_rule
            if tenant_id == tenant:
                if tenant_rule:
                    raise TenantPolicyException(
                        error_msg='Number of tenant specific router rules '
                                  'exceeds the limit of 1!')
                tenant_rule = parsed_rule
        return tenant_rule if tenant_rule else default_rule

    def create_default_policy(self, context, tenant_id, default_policy_dict):
        with db_api.CONTEXT_WRITER.using(context):
            existing_def_rule = (context.session.query(TenantPolicy)
                                 .filter_by(tenant_id=tenant_id)
                                 .filter_by(priority=DEFAULT_POLICY_PRIORITY)
                                 .one_or_none())
            if existing_def_rule:
                return None
            default_policy = TenantPolicy(
                id=uuidutils.generate_uuid(),
                tenant_id=tenant_id,
                priority=DEFAULT_POLICY_PRIORITY,
                source=default_policy_dict['source'],
                source_port=None,
                destination=default_policy_dict['destination'],
                destination_port=None,
                protocol=None,
                action=default_policy_dict['action'],
                nexthops=default_policy_dict['nexthops'])
            context.session.add(default_policy)
        return self._make_tenantpolicy_dict(default_policy)

    def remove_default_policy(self, context, tenant_id):
        with db_api.CONTEXT_WRITER.using(context):
            # tenant doesn't have another router, remove all policies
            LOG.debug("Tenant doesn't have another router after router "
                      "deletion. Removing all policies under the tenant.")
            (context.session.query(TenantPolicy)
             .filter_by(tenant_id=tenant_id)
             .delete(synchronize_session='fetch'))
            context.session.flush()
            return



# Copyright 2017, Big Switch Networks
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
"""testpath add column logical path

Revision ID: 2dc6f1b7c0a1
Revises: f774eae87144
Create Date: 2017-10-12 14:12:51.467479

"""

import sqlalchemy as sa

from alembic import op
from oslo_serialization import jsonutils
from sqlalchemy.dialects.mysql.base import VARCHAR
from sqlalchemy.types import TypeDecorator

# revision identifiers, used by Alembic.
revision = '2dc6f1b7c0a1'
down_revision = 'f774eae87144'
branch_labels = None
depends_on = None


class JSONEncodedDict(TypeDecorator):
    """Represents an immutable structure as a json-encoded string.

    Usage::
        JSONEncodedDict(255)
    """
    impl = VARCHAR

    def process_bind_param(self, value, dialect):
        if value is not None:
            value = jsonutils.dumps(value)
        return value

    def process_result_value(self, value, dialect):
        if value is not None:
            value = jsonutils.loads(value)
        return value


def upgrade():
    op.add_column('reachabilitytest',
                  sa.Column('logical_path', JSONEncodedDict(8192),
                            nullable=True))
    op.add_column('reachabilityquicktest',
                  sa.Column('logical_path', JSONEncodedDict(8192),
                            nullable=True))


def downgrade():
    pass



# Copyright 2011 OpenStack Foundation.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import sqlalchemy as sa

from networking_bigswitch.plugins.bigswitch.i18n import _
from neutron_lib.db import model_base
from neutron_lib.db import model_query
from neutron_lib.db import utils as db_utils
from neutron_lib import exceptions
from oslo_db import exception as db_exc
from sqlalchemy.orm import exc


class NetworkTemplate(model_base.BASEV2, model_base.HasId):
    __tablename__ = 'networktemplates'
    __table_args__ = {'extend_existing': True}
    body = sa.Column(sa.Text(), nullable=False)
    name = sa.Column(sa.String(255), nullable=False, unique=True)


class NetworkTemplateNotFound(exceptions.NotFound):
    message = _("Network Template %(id)s could not be found")


class NetworkTemplateDbMixin(object):
    # internal methods
    def _make_networktemplate_dict(self, template, fields=None):
        return db_utils.resource_fields({
            'id': template.id,
            'body': template.body,
            'name': template.name}, fields)

    def _get_networktemplate(self, context, id):
        try:
            networktemplate = model_query.get_by_id(
                context, NetworkTemplate, id)
        except exc.NoResultFound:
            raise NetworkTemplateNotFound(id=id)
        return networktemplate

    # public CRUD methods for network templates
    def get_networktemplates(self, context, filters=None, fields=None,
                             sorts=None, limit=None, marker=None,
                             page_reverse=False):
        networktemplates = \
            model_query.get_collection(context, NetworkTemplate,
                                       self._make_networktemplate_dict,
                                       filters=filters, fields=fields)
        return networktemplates

    def get_networktemplate(self, context, id, fields=None):
        networktemplate = self._get_networktemplate(context, id)
        return self._make_networktemplate_dict(networktemplate, fields)

    def create_networktemplate(self, context, networktemplate):
        networktemplate_data = networktemplate['networktemplate']
        with context.session.begin(subtransactions=True):
            networktemplate = \
                NetworkTemplate(body=networktemplate_data['body'],
                                name=networktemplate_data['name'])
            context.session.add(networktemplate)
        return self._make_networktemplate_dict(networktemplate)

    def delete_networktemplate(self, context, id):
        with context.session.begin(subtransactions=True):
            networktemplate = self._get_networktemplate(context, id)
            context.session.delete(networktemplate)

    def update_networktemplate(self, context, id, networktemplate):
        networktemplate_data = networktemplate['networktemplate']
        with context.session.begin(subtransactions=True):
            networktemplate = self._get_networktemplate(context, id)
            networktemplate.update(networktemplate_data)
        return self._make_networktemplate_dict(networktemplate)


class NetworkTemplateAssignment(model_base.BASEV2,
                                model_base.HasId,
                                model_base.HasProject):
    __tablename__ = 'networktemplateassignments'
    __table_args__ = {'extend_existing': True}
    template_id = sa.Column(sa.Integer, sa.ForeignKey('networktemplates.id'),
                            nullable=False)
    stack_id = sa.Column(sa.String(255), nullable=False)
    template = sa.orm.relationship("NetworkTemplate")


class NetworkTemplateAssignmentNotFound(exceptions.NotFound):
    message = _("Network Template Assignment %(id)s could not be found")


class NetworkTemplateAssignmentExists(exceptions.NeutronException):
    message = \
        _("Network Template Assignment for tenant ID %(tenant_id)s exists")


class NetworkTemplateAssignmentDbMixin(object):
    # internal methods
    def _make_networktemplateassignment_dict(self, templateassignment,
                                             fields=None):
        return db_utils.resource_fields({
            'id': templateassignment.id,
            'template_id': templateassignment.template_id,
            'tenant_id': templateassignment.tenant_id,
            'stack_id': templateassignment.stack_id}, fields)

    def _get_networktemplateassignment(self, context, id):
        try:
            networktemplateassignment = model_query.get_by_id(
                context, NetworkTemplateAssignment, id)
        except exc.NoResultFound:
            raise NetworkTemplateAssignmentNotFound(id=id)
        return networktemplateassignment

    # public CRUD methods for Network Template Assignment
    def get_networktemplateassignments(self, context, filters=None,
                                       fields=None, sorts=None, limit=None,
                                       marker=None, page_reverse=False):
        networktemplateassignments = \
            model_query.get_collection(
                context, NetworkTemplateAssignment,
                self._make_networktemplateassignment_dict,
                filters=filters, fields=fields)
        return networktemplateassignments

    def get_networktemplateassignment(self, context, id, fields=None):
        networktemplateassignment = \
            self._get_networktemplateassignment(context, id)
        return self._make_networktemplateassignment_dict(
            networktemplateassignment, fields)

    def create_networktemplateassignment(self, context,
                                         networktemplateassignment):
        networktemplateassignment_data = \
            networktemplateassignment['networktemplateassignment']
        with context.session.begin(subtransactions=True):
            networktemplateassignment = NetworkTemplateAssignment(
                id=networktemplateassignment_data['tenant_id'],
                tenant_id=networktemplateassignment_data['tenant_id'],
                template_id=networktemplateassignment_data['template_id'],
                stack_id=networktemplateassignment_data['stack_id'])
            try:
                context.session.add(networktemplateassignment)
            except db_exc.DBDuplicateEntry:
                raise NetworkTemplateAssignmentExists(
                    tenant_id=networktemplateassignment.tenant_id)
        return self._make_networktemplateassignment_dict(
            networktemplateassignment)

    def delete_networktemplateassignment(self, context, id):
        with context.session.begin(subtransactions=True):
            networktemplateassignment = \
                self._get_networktemplateassignment(context, id)
            context.session.delete(networktemplateassignment)

    def update_networktemplateassignment(self, context, id,
                                         networktemplateassignment):
        networktemplateassignment_data = \
            networktemplateassignment['networktemplateassignment']
        with context.session.begin(subtransactions=True):
            networktemplateassignment = \
                self._get_networktemplateassignment(context, id)
            networktemplateassignment.update(networktemplateassignment_data)
        return self._make_networktemplateassignment_dict(
            networktemplateassignment)



# Copyright 2014, Big Switch Networks
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
import datetime
import eventlet
import random
import re
import time

from oslo_config import cfg
from oslo_db import exception as db_exc
from oslo_db.sqlalchemy import session
from oslo_log import log as logging
import sqlalchemy as sa

from neutron_lib.db import model_base
from neutron_lib import exceptions

from networking_bigswitch.plugins.bigswitch.i18n import _

LOG = logging.getLogger(__name__)

MIN_LOCK_RETRY_SLEEP_TIME_SECS = 5
MAX_LOCK_RETRY_SLEEP_TIME_SECS = 30
MAX_LOCK_RETRY_COUNT = 12
TOPO_SYNC_EXPIRED_SECS = 1800


class ConsistencyHash(model_base.BASEV2):
    """Consistency Hash

    A simple table to store the latest consistency hash
    received from a server.
    For now we only support one global state so the
    hash_id will always be '1'
    """
    __tablename__ = 'consistencyhashes'
    hash_id = sa.Column(sa.String(255),
                        primary_key=True)
    hash = sa.Column(sa.String(255), nullable=False)


def setup_db():
    '''Helper to register models for unit tests'''
    if HashHandler._FACADE is None:
        HashHandler._FACADE = session.EngineFacade.from_config(
            cfg.CONF, sqlite_fk=True)
    ConsistencyHash.metadata.create_all(
        HashHandler._FACADE.get_engine())


def clear_db():
    '''Helper to unregister models and clear engine in unit tests'''
    if not HashHandler._FACADE:
        return
    ConsistencyHash.metadata.drop_all(HashHandler._FACADE.get_engine())
    HashHandler._FACADE = None


def convert_ts_to_datetime(timestamp_ms):
    dt_string = datetime.datetime.fromtimestamp(float(timestamp_ms)).strftime(
        '%Y-%m-%d %H:%M:%S.%f')
    return dt_string


def log_lock_acquisition_failure(prev_ts, curr_ts):
    LOG.debug(
        "TOPO_SYNC: Failed to acquire lock. Restarting lock wait. "
        "PreviousTS: %(prev_ts)s, datetime %(prev_dt_string)s. "
        "Attempted update: %(curr_ts)s, datetime %(curr_dt_string)s.",
        {'prev_ts': prev_ts, 'prev_dt_string': convert_ts_to_datetime(prev_ts),
         'curr_ts': curr_ts, 'curr_dt_string': convert_ts_to_datetime(curr_ts)
         })


def get_lock_owner(hash):
    matches = re.findall(r"^TOPO_SYNC\[(\d+\.\d+)\]", hash)
    if not matches:
        return None
    return matches[0]


class LockRetryCountExceededException(exceptions.NeutronException):
    message = _("TOPO_SYNC: LockTS %(lock_ts)s, datetime %(lock_dt_string)s "
                "exceeded MAX_LOCK_RETRY_COUNT %(max_lock)s.")
    status = None

    def __init__(self, **kwargs):
        self.lock_ts = kwargs.get('lock_ts')
        self.lock_dt_string = kwargs.get('lock_dt_string')
        self.max_lock = kwargs.get('max_lock')
        super(LockRetryCountExceededException, self).__init__(**kwargs)


class HashHandler(object):
    """Hash Handler

    A wrapper object to keep track of the session between the read
    and the update operations.

    This class needs an SQL engine completely independent of the main
    neutron connection so rollbacks from consistency hash operations don't
    affect the parent sessions.
    """
    _FACADE = None

    def __init__(self, hash_id='1', timestamp_ms=None):
        if HashHandler._FACADE is None:
            HashHandler._FACADE = session.EngineFacade.from_config(
                cfg.CONF, sqlite_fk=True)

        self.hash_id = hash_id
        self.session = HashHandler._FACADE.get_session(autocommit=True,
                                                       expire_on_commit=False)
        self.lock_ts = str(timestamp_ms) if timestamp_ms else str(time.time())
        self.lock_marker = 'TOPO_SYNC[%s]' % self.lock_ts
        self.lock_retry_count = 0
        # before grabbing the lock, store the previous timestamp
        # in case there is an exception while collecting data or updating BCF,
        # revert to previous timestamp
        self.prev_lock_ts = '0'

    def _increment_lock_retry(self):
        """Increments lock retry count.

        Raises exception when it goes over the retry limit.

        :return:
        """
        self.lock_retry_count += 1
        if self.lock_retry_count >= MAX_LOCK_RETRY_COUNT:
            raise LockRetryCountExceededException(
                lock_ts=self.lock_ts,
                lock_dt_string=convert_ts_to_datetime(self.lock_ts),
                max_lock=MAX_LOCK_RETRY_COUNT)

    def _get_current_record(self):
        with self.session.begin(subtransactions=True):
            res = (self.session.query(ConsistencyHash).
                   filter_by(hash_id=self.hash_id).first())
            if res:
                self.session.refresh(res)  # make sure latest is loaded from db
            return res

    def _insert_hash_with_lock(self):
        # try to insert a new hash, return False on conflict
        try:
            with self.session.begin(subtransactions=True):
                res = ConsistencyHash(hash_id=self.hash_id,
                                      hash=self.lock_marker)
                self.session.add(res)
                return True
        except db_exc.DBDuplicateEntry:
            # another server created a new record at the same time
            return False

    def _optimistic_update_hash_record(self, old_record, new_hash):
        # Optimistic update strategy. Returns True if successful, else False.
        query = sa.update(ConsistencyHash.__table__).values(hash=new_hash)
        query = query.where(ConsistencyHash.hash_id == old_record.hash_id)
        query = query.where(ConsistencyHash.hash == old_record.hash)
        try:
            with self._FACADE.get_engine().begin() as conn:
                result = conn.execute(query)
        except db_exc.DBDeadlock:
            # mysql can encounter internal deadlocks servicing a query with
            # multiple where criteria. treat it the same as not being able
            # to update the record so it will be tried again
            return False
        # We need to check update row count in case another server is
        # doing this at the same time. Only one will succeed, the other will
        # not update any rows.
        return result.rowcount != 0

    def _is_timestamp_expired(self, expire_secs=TOPO_SYNC_EXPIRED_SECS,
                              prev_ts='0'):
        """Check if prev_ts is older than TOPO_SYNC_EXPIRED_SECS

        :param prev_ts: previous timestamp value from hash record or 0
        :return: (bool, prev_ts)
                 boolean True if prev_ts is < TOPO_SYNC_EXPIRED_SECS else False
        """
        # current thread checking for time diff may have returned after sleep,
        # we check time as of NOW, not the timestamp associated with the lock
        # checking for expiration of previous one.
        curr_ts_float = time.time()
        try:
            prev_ts_float = float(prev_ts)
        except Exception:
            # prev_ts value was garbage, set it to 0
            prev_ts = '0'
            prev_ts_float = float(prev_ts)

        time_diff = curr_ts_float - prev_ts_float
        if time_diff < expire_secs:
            # diff is under the limit, NOT EXPIRED
            LOG.debug("TOPO_SYNC: NOT EXPIRED. "
                      "PreviousTS %(prev)s, datetime %(prev_dt_string)s. "
                      "LockTS %(curr)s, datetime %(curr_dt_string)s. "
                      "Diff < %(expired_secs)s secs.",
                      {'prev': prev_ts,
                       'prev_dt_string': convert_ts_to_datetime(prev_ts_float),
                       'curr': self.lock_ts,
                       'curr_dt_string': convert_ts_to_datetime(self.lock_ts),
                       'diff': time_diff,
                       'expired_secs': expire_secs})
            return False
        else:
            # diff is over the time range, EXPIRED
            LOG.debug("TOPO_SYNC: EXPIRED. "
                      "PreviousTS %(prev)s, datetime %(prev_dt_string)s. "
                      "LockTS %(curr)s, datetime %(curr_dt_string)s. "
                      "Diff < %(expired_secs)s secs.",
                      {'prev': prev_ts,
                       'prev_dt_string': convert_ts_to_datetime(prev_ts_float),
                       'curr': self.lock_ts,
                       'curr_dt_string': convert_ts_to_datetime(self.lock_ts),
                       'diff': time_diff,
                       'expired_secs': expire_secs})
            return True

    def lock(self, check_ts=True):
        """Lock based on the below condition:

        while -
            if no hash present
              insert lock and move on

            if hash present
              - check if TOPO_SYNC is present i.e. TopoSync ongoing or expired
                 - (y) check if the Previous TopoSync has expired
                    - (y) evict it, put self ts and move on
                  - if locked by self, return
                  - else if check_ts is false, sleep_and_retry

              - if no TOPO_SYNC
                - grab lock, check if TS has expired
                  - if expired, return true
                  - if not, put back old TS and return false

        :param check_ts:
        :return: boolean sync_required
                            True if all conditions met and lock acquired
                            False if locking not required
        """
        retry_sleep_time = random.randint(MIN_LOCK_RETRY_SLEEP_TIME_SECS,
                                          MAX_LOCK_RETRY_SLEEP_TIME_SECS)
        new_hash = self.lock_marker
        while True:
            res = self._get_current_record()
            # set prev_lock_ts based on record
            if res:
                if 'TOPO_SYNC' in res.hash:
                    self.prev_lock_ts = get_lock_owner(res.hash)
                else:
                    self.prev_lock_ts = str(res.hash)

            # try lock acquisition based on hash record
            if not res:
                # no hash present, try optimistically locking it
                if not self._insert_hash_with_lock():
                    # someone else beat us to it, sleep and retry
                    log_lock_acquisition_failure('0', self.lock_ts)
                    eventlet.sleep(retry_sleep_time)
                    continue
                # got the lock, execute update since nothing existed earlier
                LOG.debug(
                    "TOPO_SYNC: LockTS %(lockts)s, datetime %(dt_string)s has "
                    "grabbed the lock.",
                    {'lockts': self.lock_ts,
                     'dt_string': convert_ts_to_datetime(self.lock_ts)})
                return True

            if 'TOPO_SYNC' in res.hash:
                # TOPO_SYNC already in progress. Check if timestamp is over
                # the limit
                prev_ts = get_lock_owner(res.hash)
                if self._is_timestamp_expired(
                        expire_secs=TOPO_SYNC_EXPIRED_SECS, prev_ts=prev_ts):
                    # optimistically update timestamp
                    if not self._optimistic_update_hash_record(res, new_hash):
                        # someone else update it before us, return
                        log_lock_acquisition_failure(prev_ts, self.lock_ts)
                        eventlet.sleep(retry_sleep_time)
                        continue
                    # grabbed the lock
                    LOG.debug(
                        "TOPO_SYNC: LockTS %(lock_ts)s, datetime "
                        "%(lock_dt_string)s has forcefully grabbed the lock. "
                        "PreviousTS %(prev_ts)s, datetime %(prev_dt_string)s "
                        "was over %(expired_secs)s old.",
                        {'lock_ts': self.lock_ts,
                         'lock_dt_string': convert_ts_to_datetime(
                             self.lock_ts),
                         'prev_ts': prev_ts,
                         'prev_dt_string': convert_ts_to_datetime(prev_ts)})
                    return True
                else:
                    if prev_ts == self.lock_ts:
                        LOG.debug("TOPO_SYNC: LockTS %(lockts)s, datetime "
                                  "%(dt_string)s has grabbed the lock.",
                                  {'lockts': self.lock_ts,
                                   'dt_string': convert_ts_to_datetime(
                                       self.lock_ts)})
                        return True

                    if check_ts:
                        LOG.debug(
                            "TOPO_SYNC: LockTS %(lock_ts)s, datetime "
                            "%(lock_dt_string)s giving up since previous lock "
                            "not expired.",
                            {'lock_ts': self.lock_ts, 'lock_dt_string':
                                convert_ts_to_datetime(self.lock_ts)})
                        return False
                    LOG.debug(
                        "TOPO_SYNC: LockTS %(lock_ts)s, datetime "
                        "%(lock_dt_string)s waiting for in progress topo_sync "
                        "to complete.",
                        {'lock_ts': self.lock_ts, 'lock_dt_string':
                            convert_ts_to_datetime(self.lock_ts)})
                    # this is the only place where we retry while waiting for
                    # timeout. don't wait indefinitely
                    self._increment_lock_retry()
                    eventlet.sleep(retry_sleep_time)
                    continue
            else:
                # nobody has the lock, grab it!
                if not self._optimistic_update_hash_record(res, new_hash):
                    # someone else updated it before us, sleep and retry
                    log_lock_acquisition_failure(res.hash, self.lock_ts)
                    eventlet.sleep(retry_sleep_time)
                    continue

                # finally grabbed the lock
                LOG.debug(
                    "TOPO_SYNC: LockTS %(lockts)s, datetime %(dt_string)s has "
                    "grabbed the lock.",
                    {'lockts': self.lock_ts,
                     'dt_string': convert_ts_to_datetime(self.lock_ts)})
                if check_ts and not self._is_timestamp_expired(
                        expire_secs=TOPO_SYNC_EXPIRED_SECS, prev_ts=res.hash):
                    # replace with old hash, since we already grabbed the lock
                    LOG.debug(
                        "TOPO_SYNC: Giving up lock since check_ts is True and "
                        "previous timestamp not expired.")
                    self.put_hash(res.hash)
                    return False
                # lock grabbed and not returned. return True
                return True

    def put_hash(self, new_hash):
        query = sa.update(ConsistencyHash.__table__).values(hash=new_hash)
        query = query.where(ConsistencyHash.hash_id == self.hash_id)

        try:
            with self._FACADE.get_engine().begin() as conn:
                conn.execute(query)
        except db_exc.DBDeadlock:
            LOG.debug("TOPO_SYNC: Failed to update timestamp to previous "
                      "value %(new_hash)s by LockTS %(this)s.",
                      {'new_hash': new_hash, 'this': self.lock_ts})

        # DB Hash update was successful
        LOG.debug("TOPO_SYNC: Consistency timestamp updated to previous value "
                  "%(hash_ts)s by LockID %(this)s",
                  {'hash_ts': new_hash, 'this': self.lock_ts})
        return

    def unlock(self, set_prev_ts=False):
        """Unlock the consistency DB record if locked by TOPO_SYNC

        If TOPO_SYNC not found in hash record, it was unlocked by some other
        thread, do nothing.

        If set_prev_ts is True, unlock and set the timestamp to prev_lock_ts.
        This is in cases when topo_sync has failed.

        :param set_prev_ts:
        :return:
        """
        unlock_ts = self.lock_ts
        if set_prev_ts:
            unlock_ts = self.prev_lock_ts
        LOG.debug("TOPO_SYNC: Unlocking and setting LockTS  to %s",
                  unlock_ts)
        with self.session.begin(subtransactions=True):
            res = (self.session.query(ConsistencyHash).
                   filter_by(hash_id=self.hash_id).first())
            if not res:
                LOG.warning("Hash record already gone, no lock to clear.")
                return
            else:
                self.session.refresh(res)  # get the latest res from db
            if res.hash != self.lock_marker:
                # if these are frequent the server is too slow
                LOG.warning("Another server already removed the lock. %s",
                            res.hash)
                return
            res.hash = res.hash.replace(self.lock_marker, unlock_ts)



# Copyright 2013 Big Switch Networks, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from neutron.tests.unit.db import test_agentschedulers_db

from networking_bigswitch.tests.unit.bigswitch import test_base


class BigSwitchDhcpAgentNotifierTestCase(
        test_agentschedulers_db.OvsDhcpAgentNotifierTestCase,
        test_base.BigSwitchTestBase):

    plugin_str = ('%s.NeutronRestProxyV2' %
                  test_base.PLUGIN_PATH)

    def setUp(self):
        self.setup_config_files()
        self.setup_patches()
        test_agentschedulers_db.OvsDhcpAgentNotifierTestCase.setUp(self)
        self.setup_db()
        self.startHttpPatch()



# Copyright 2014 Big Switch Networks, Inc.
# All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import mock
from neutron.conf.plugins.ml2 import config as ml2_config
from neutron.plugins.ml2.drivers import type_vlan as vlan_config
from neutron.tests.unit.api import test_extensions
from neutron.tests.unit.db import test_db_base_plugin_v2
from neutron.tests.unit.extensions import test_securitygroup as test_sg
from neutron.tests.unit.plugins.ml2 import test_plugin
from neutron_lib.api.definitions import portbindings
from neutron_lib import context as neutron_context
from neutron_lib.plugins import directory
from oslo_serialization import jsonutils
from oslo_utils import uuidutils

from networking_bigswitch.plugins.bigswitch import servermanager
from networking_bigswitch.tests.unit.bigswitch.fake_server \
    import HTTPResponseMock
from networking_bigswitch.tests.unit.bigswitch.mock_paths import DRIVER
from networking_bigswitch.tests.unit.bigswitch.mock_paths import HTTPCON
from networking_bigswitch.tests.unit.bigswitch.mock_paths import SERVER_POOL
import networking_bigswitch.tests.unit.bigswitch.test_restproxy_plugin as trp

_uuid = uuidutils.generate_uuid

PHYS_NET = 'physnet1'
VLAN_START = 1000
VLAN_END = 1100


class TestBigSwitchMechDriverBase(trp.BigSwitchProxyPluginV2TestCase):

    def setUp(self, plugin=None, service_plugins=None, ext_mgr=None):
        # Configure the ML2 mechanism drivers and network types
        ml2_opts = {
            'mechanism_drivers': ['bsn_ml2'],
            'tenant_network_types': ['vlan'],
        }
        for opt, val in ml2_opts.items():
                ml2_config.cfg.CONF.set_override(opt, val, 'ml2')

        # Configure the ML2 VLAN parameters
        phys_vrange = ':'.join([PHYS_NET, str(VLAN_START), str(VLAN_END)])
        vlan_config.cfg.CONF.set_override('network_vlan_ranges',
                                          [phys_vrange],
                                          'ml2_type_vlan')
        super(TestBigSwitchMechDriverBase,
              self).setUp(test_plugin.PLUGIN_NAME,
                          service_plugins=service_plugins,
                          ext_mgr=ext_mgr)


class TestBigSwitchMechDriverNetworksV2(test_db_base_plugin_v2.TestNetworksV2,
                                        TestBigSwitchMechDriverBase):
    def setUp(self, plugin=None, service_plugins=None, ext_mgr=None):
        TestBigSwitchMechDriverBase.setUp(self,
                                          plugin=plugin,
                                          service_plugins=service_plugins,
                                          ext_mgr=ext_mgr)

    def test_create_network(self):
        name = 'net1'
        keys = [('subnets', []), ('name', name), ('admin_state_up', True),
                ('status', self.net_create_status), ('shared', False)]

        with mock.patch(HTTPCON) as conmock:
            rv = conmock.return_value
            rv.getresponse.return_value = HTTPResponseMock(None)
            with self.network(name=name) as net:
                # for debug
                print(rv.request.mock_calls)

                # remove topo sync call
                filtered_calls = \
                    list(filter(lambda call: "topology" not in call[1][1],
                                rv.request.mock_calls))

                network = jsonutils.loads(filtered_calls[0][1][2])
                self.assertIn('tenant_name', network['network'])
                self.assertEqual('tenant_name',
                                 network['network']['tenant_name'])
                for k, v in keys:
                    self.assertEqual(net['network'][k], v)

    def test_update_network(self):
        with self.network() as network:
            data = {'network': {'name': 'a_brand_new_name'}}
            req = self.new_update_request('networks',
                                          data,
                                          network['network']['id'])
            res = self.deserialize(self.fmt, req.get_response(self.api))
            self.assertTrue('NeutronError' in res)
            self.assertEqual('NetworkNameChangeError',
                             res['NeutronError']['type'])


class TestBigSwitchML2SubnetsV2(test_db_base_plugin_v2.TestSubnetsV2,
                                TestBigSwitchMechDriverBase):
    pass


class TestBigSwitchML2SecurityGroups(test_sg.TestSecurityGroups,
                                     TestBigSwitchMechDriverBase):
    def setUp(self):
        ext_mgr = test_sg.SecurityGroupTestExtensionManager()
        super(TestBigSwitchML2SecurityGroups, self).setUp(ext_mgr=ext_mgr)
        self.ext_api = test_extensions.setup_extensions_middleware(ext_mgr)


class TestBigSwitchMechDriverPortsV2(test_db_base_plugin_v2.TestPortsV2,
                                     TestBigSwitchMechDriverBase):

    VIF_TYPE = portbindings.VIF_TYPE_OVS

    def setUp(self):
        super(TestBigSwitchMechDriverPortsV2, self).setUp()
        self.port_create_status = 'DOWN'

    def test_update_port_status_build(self):
        with self.port() as port:
            self.assertEqual(port['port']['status'], 'DOWN')
            self.assertEqual(self.port_create_status, 'DOWN')

    def test_dont_bind_vnic_type_direct(self):
        host_arg = {portbindings.HOST_ID: 'hostname',
                    portbindings.VNIC_TYPE: portbindings.VNIC_DIRECT}
        with\
            mock.patch(SERVER_POOL + '.rest_get_switch',
                       return_value=True) as rmock,\
            self.port(arg_list=(portbindings.HOST_ID, portbindings.VNIC_TYPE),
                      **host_arg):

            # bind_port() shall ignore this call
            rmock.assert_not_called()

    def test_dont_bind_vnic_type_direct_physical(self):
        host_arg = {portbindings.HOST_ID: 'hostname',
                    portbindings.VNIC_TYPE: portbindings.VNIC_DIRECT_PHYSICAL}
        with\
            mock.patch(SERVER_POOL + '.rest_get_switch',
                       return_value=True) as rmock,\
            self.port(arg_list=(portbindings.HOST_ID, portbindings.VNIC_TYPE),
                      **host_arg):

            # bind_port() shall ignore this call
            rmock.assert_not_called()

    def test_create404_triggers_background_sync(self):
        # allow the async background thread to run for this test
        self.spawn_p.stop()
        with\
            mock.patch(
                SERVER_POOL + '.rest_create_port',
                side_effect=servermanager.RemoteRestError(
                    reason=servermanager.NXNETWORK, status=404)),\
            mock.patch(DRIVER + '._send_all_data') as mock_send_all,\
            self.port(**{'device_id': 'devid', 'binding:host_id': 'host',
                         'arg_list': ('binding:host_id',)}) as p:

            # wait for thread to finish
            mm = directory.get_plugin().mechanism_manager
            bigdriver = mm.mech_drivers['bsn_ml2'].obj
            bigdriver.evpool.waitall()
            mock_send_all.assert_has_calls([
                mock.call(
                    timeout=None,
                    triggered_by_tenant=p['port']['tenant_id']
                )
            ])
        self.spawn_p.start()

    def test_udpate404_triggers_background_sync(self):
        with mock.patch(DRIVER + '.async_port_create',
                        side_effect=servermanager.RemoteRestError(
                            reason=servermanager.NXNETWORK,
                            status=404)),\
                mock.patch(DRIVER + '._send_all_data') as mock_send_all,\
                self.port() as p:

            plugin = directory.get_plugin()
            context = neutron_context.get_admin_context()
            plugin.update_port(context, p['port']['id'],
                               {'port': {'device_id': 'devid',
                                         'binding:host_id': 'host'}})
            mock_send_all.assert_has_calls([
                mock.call(
                    timeout=None,
                    triggered_by_tenant=p['port']['tenant_id']
                )
            ])

    def test_backend_request_contents(self):
        with\
            mock.patch(SERVER_POOL + '.rest_create_port') as mock_rest,\
            self.port(**{'device_id': 'devid', 'binding:host_id': 'host',
                         'arg_list': ('binding:host_id',)}):

            # make sure basic expected keys are present in the port body
            pb = mock_rest.mock_calls[0][1][2]
            self.assertEqual('host', pb['binding:host_id'])
            self.assertIn('bound_segment', pb)
            self.assertIn('network', pb)



#!/usr/bin/python
# Copyright 2018 Big Switch Networks, Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#

import datetime
import requests
import sys

# read arguments
if len(sys.argv) != 5:
    sys.exit("Received incorrect number of arguments to script, 4 expected.")

conf_user = sys.argv[1]
conf_token = sys.argv[2]
page_id = sys.argv[3]
html_href = sys.argv[4]
print("page_id is: %(page_id)s" % {'page_id': page_id})
print("html_href is: %(html_href)s" % {'html_href': html_href})
basic_auth = (conf_user, conf_token)
print("basic_auth is %(auth)s" % {'auth': basic_auth})

headers = {'content-type': 'application/json'}

# get newton existing page with body and version number
url = ('https://bigswitch.atlassian.net/wiki/rest/api/content/%(page_id)s'
       '?expand=body.storage,version' % {'page_id': page_id})
r = requests.get(url, headers=headers, auth=basic_auth)
json_out = r.json()
version_number = json_out['version']['number']
new_ver_num = version_number + 1

datetime_str = datetime.datetime.now().strftime("%B %d, %Y at %I:%M%p PST")
new_value = html_href + '<br/>' + datetime_str

# create new data to be uploaded
data = ('{"id":"%(page_id)s",'
        '"body":{"storage":{"value": "%(new_value)s",'
        '"representation":"storage"}},'
        '"version": {"number": %(new_ver_num)s},'
        '"type":"%(type)s","title":"%(title)s"}'
        % {'page_id': page_id,
           'new_value': new_value,
           'new_ver_num': new_ver_num,
           'type': json_out['type'],
           'title': json_out['title']})

p = requests.put(url, data=data, headers=headers, auth=basic_auth)
print(p.json())



# Copyright 2013 Big Switch Networks, Inc.  All rights reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from oslo_log import log as logging
from oslo_serialization import jsonutils

from networking_bigswitch.plugins.bigswitch.i18n import _
from networking_bigswitch.plugins.bigswitch import servermanager

LOG = logging.getLogger(__name__)


class HTTPResponseMock(object):
    status = 200
    reason = 'OK'

    def __init__(self, sock, debuglevel=0, strict=0, method=None,
                 buffering=False):
        pass

    def read(self):
        return "{'status': '200 OK'}"

    def getheader(self, header):
        return None


class HTTPResponseMock404(HTTPResponseMock):
    status = 404
    reason = 'Not Found'

    def read(self):
        return "{'status': '%s 404 Not Found'}" % servermanager.NXNETWORK


class HTTPResponseMock500(HTTPResponseMock):
    status = 500
    reason = 'Internal Server Error'

    def __init__(self, sock, debuglevel=0, strict=0, method=None,
                 buffering=False, errmsg='500 Internal Server Error'):
        self.errmsg = errmsg

    def read(self):
        return "{'status': '%s'}" % self.errmsg


class HTTPConnectionMock(object):

    def __init__(self, host, port, timeout):
        self.response = None
        self.broken = False
        # Port 9000 is the broken server
        if port == 9000:
            self.broken = True
            errmsg = "This server is broken, please try another"
            self.response = HTTPResponseMock500(None, errmsg=errmsg)

    def request(self, action, uri, body, headers):
        LOG.debug("Request: action=%(action)s, uri=%(uri)r, "
                  "body=%(body)s, headers=%(headers)s",
                  {'action': action, 'uri': uri,
                   'body': body, 'headers': headers})
        if self.broken and "ExceptOnBadServer" in uri:
            raise Exception("Broken server got an unexpected request")
        if self.response:
            return

        # detachment may return 404 and plugin shouldn't die
        if uri.endswith('attachment') and action == 'DELETE':
            self.response = HTTPResponseMock404(None)
        else:
            self.response = HTTPResponseMock(None)

        # Port creations/updates must contain binding information
        if ('port' in uri and 'attachment' not in uri and
                'binding' not in body and action in ('POST', 'PUT')):
            errmsg = "Port binding info missing in port request '%s'" % body
            self.response = HTTPResponseMock500(None, errmsg=errmsg)
            return

        return

    def getresponse(self):
        return self.response

    def close(self):
        pass


class HTTPConnectionMock404(HTTPConnectionMock):

    def __init__(self, host, port, timeout):
        self.response = HTTPResponseMock404(None)
        self.broken = True


class HTTPConnectionMock500(HTTPConnectionMock):

    def __init__(self, host, port, timeout):
        self.response = HTTPResponseMock500(None)
        self.broken = True


class VerifyMultiTenantFloatingIP(HTTPConnectionMock):

    def request(self, action, uri, body, headers):
        # Only handle network update requests
        if 'network' in uri and 'tenant' in uri and 'ports' not in uri:
            req = jsonutils.loads(body)
            if 'network' not in req or 'floatingips' not in req['network']:
                msg = _("No floating IPs in request"
                        "uri=%(uri)s, body=%(body)s") % {'uri': uri,
                                                         'body': body}
                raise Exception(msg)
            distinct_tenants = []
            for flip in req['network']['floatingips']:
                if flip['tenant_id'] not in distinct_tenants:
                    distinct_tenants.append(flip['tenant_id'])
            if len(distinct_tenants) < 2:
                msg = _("Expected floating IPs from multiple tenants."
                        "uri=%(uri)s, body=%(body)s") % {'uri': uri,
                                                         'body': body}
                raise Exception(msg)
        super(VerifyMultiTenantFloatingIP,
              self).request(action, uri, body, headers)


class HTTPSMockBase(HTTPConnectionMock):
    def __init__(self, host, port=None, context=None, timeout=None,
                 source_address=None):
        self.host = host
        self.ssl_context = context
        super(HTTPSMockBase, self).__init__(host, port, timeout)



# Copyright 2013, Big Switch Networks
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from neutron_lib.api import validators
from oslo_log import log as logging

LOG = logging.getLogger(__name__)


def get_port_hostid(context, port_id):
    # REVISIT(kevinbenton): this is a workaround to avoid portbindings_db
    # relational table generation until one of the functions is called.
    from neutron.db.models import portbinding
    with context.session.begin(subtransactions=True):
        query = context.session.query(portbinding.PortBindingPort)
        res = query.filter_by(port_id=port_id).first()
    if not res:
        return False
    return res.host


def put_port_hostid(context, port_id, host):
    # REVISIT(kevinbenton): this is a workaround to avoid portbindings_db
    # relational table generation until one of the functions is called.
    from neutron.db.models import portbinding
    if not validators.is_attr_set(host):
        LOG.warning("No host_id in port request to track port location.")
        return
    if port_id == '':
        LOG.warning("Received an empty port ID for host_id '%s'", host)
        return
    if host == '':
        LOG.debug("Received an empty host_id for port '%s'", port_id)
        return
    LOG.debug("Logging port %(port)s on host_id %(host)s",
              {'port': port_id, 'host': host})
    with context.session.begin(subtransactions=True):
        location = portbinding.PortBindingPort(port_id=port_id, host=host)
        context.session.merge(location)



#!/usr/bin/env python
# Copyright 2012 OpenStack Foundation
# Copyright 2012, Big Switch Networks, Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""Determine version of NeutronRestProxy plugin"""
from __future__ import print_function

from networking_bigswitch.plugins.bigswitch import vcsversion


YEAR, COUNT, REVISION = vcsversion.NEUTRONRESTPROXY_VERSION


def canonical_version_string():
    return '.'.join([_f for _f in vcsversion.NEUTRONRESTPROXY_VERSION if _f])


def version_string():
    if vcsversion.FINAL:
        return canonical_version_string()
    else:
        return '%s-dev' % (canonical_version_string(),)


def vcs_version_string():
    return "%s:%s" % (vcsversion.version_info['branch_nick'],
                      vcsversion.version_info['revision_id'])


def version_string_with_vcs():
    return "%s-%s" % (canonical_version_string(), vcs_version_string())


if __name__ == "__main__":
    print(version_string_with_vcs())



# Copyright 2011 OpenStack Foundation.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""testpath modify uniq constraint

Revision ID: 1ef57200f387
Revises: 1086543dfc0f
Create Date: 2016-06-18 14:04:20.625692

"""

from alembic import op

# revision identifiers, used by Alembic.
revision = '1ef57200f387'
down_revision = '1086543dfc0f'
branch_labels = None
depends_on = None


def upgrade():
    op.drop_constraint(
        'name', 'reachabilitytest',
        type_="unique")

    op.create_unique_constraint(
        constraint_name="unique_name",
        table_name="reachabilitytest",
        columns=['name', 'tenant_id'])


def downgrade():
    pass



# Copyright 2011 OpenStack Foundation.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""add base BSN plugin

Revision ID: kilo
Revises:
Create Date: 2016-01-04 17:59:34.311932

"""

from alembic import op
from oslo_serialization import jsonutils

import sqlalchemy as sa
from sqlalchemy.dialects.mysql.base import VARCHAR
from sqlalchemy.types import Enum, TIMESTAMP, TypeDecorator

# revision identifiers, used by Alembic.
revision = 'kilo'
down_revision = None
branch_labels = None
depends_on = None


class JSONEncodedDict(TypeDecorator):
    """Represents an immutable structure as a json-encoded string.

    Usage::
        JSONEncodedDict(255)
    """
    impl = VARCHAR

    def process_bind_param(self, value, dialect):
        if value is not None:
            value = jsonutils.dumps(value)
        return value

    def process_result_value(self, value, dialect):
        if value is not None:
            value = jsonutils.loads(value)
        return value


def upgrade():
    op.create_table(
        'networktemplates',
        sa.Column('id', sa.String(length=36), primary_key=True),
        sa.Column('body', sa.Text(), nullable=False),
        sa.Column('name', sa.String(255), nullable=False, unique=True))

    op.create_table(
        'networktemplateassignments',
        sa.Column('id', sa.String(length=36), primary_key=True),
        sa.Column('tenant_id', sa.String(255), nullable=False, unique=True),
        sa.Column('stack_id', sa.String(255), nullable=False),
        sa.Column('template_id', sa.String(length=36),
                  sa.ForeignKey('networktemplates.id'),
                  nullable=False),)

    op.create_table(
        'reachabilitytest',
        sa.Column('id', sa.String(length=36), primary_key=True),
        sa.Column('tenant_id', sa.String(255), nullable=False),
        sa.Column('name', sa.String(64), nullable=False, unique=True),
        sa.Column('src_tenant_name', sa.String(255), nullable=False),
        sa.Column('src_segment_name', sa.String(255), nullable=False),
        sa.Column('src_ip', sa.String(16), nullable=False),
        sa.Column('dst_ip', sa.String(16), nullable=False),
        sa.Column('expected_result',
                  Enum("reached destination", "dropped by route",
                       "dropped by policy", "dropped due to private segment",
                       "packet in", "forwarded", "dropped", "multiple sources",
                       "unsupported", "invalid input", name="expected_result"),
                  nullable=False),
        sa.Column('test_time', TIMESTAMP(timezone=True), nullable=True),
        sa.Column('test_result', Enum("pass", "fail", "pending"),
                  nullable=False, default="pending"),
        sa.Column('detail', JSONEncodedDict(8192), nullable=True),
        sa.Column('run_test', sa.Boolean, nullable=False, default=False))

    op.create_table(
        'reachabilityquicktest',
        sa.Column('id', sa.String(length=36), primary_key=True),
        sa.Column('tenant_id', sa.String(255), nullable=False),
        sa.Column('name', sa.String(64), nullable=False, unique=True),
        sa.Column('src_tenant_name', sa.String(255), nullable=False),
        sa.Column('src_segment_name', sa.String(255), nullable=False),
        sa.Column('src_ip', sa.String(16), nullable=False),
        sa.Column('dst_ip', sa.String(16), nullable=False),
        sa.Column('expected_result',
                  Enum("reached destination", "dropped by route",
                       "dropped by policy", "dropped due to private segment",
                       "packet in", "forwarded", "dropped", "multiple sources",
                       "unsupported", "invalid input", name="expected_result"),
                  nullable=False),
        sa.Column('test_time', TIMESTAMP(timezone=True), nullable=True),
        sa.Column('test_result', Enum("pass", "fail", "pending"),
                  nullable=False, default="pending"),
        sa.Column('detail', JSONEncodedDict(8192), nullable=True),
        sa.Column('run_test', sa.Boolean, nullable=False, default=False),
        sa.Column('save_test', sa.Boolean, nullable=False, default=False))

    op.create_table(
        'bsn_routerrules',
        sa.Column('id', sa.Integer(), nullable=False),
        sa.Column('priority', sa.Integer(), nullable=False),
        sa.Column('source', sa.String(length=64), nullable=False),
        sa.Column('destination', sa.String(length=64), nullable=False),
        sa.Column('action', sa.String(length=10), nullable=False),
        sa.Column('router_id', sa.String(length=36), nullable=True),
        sa.ForeignKeyConstraint(['router_id'], ['routers.id'],
                                ondelete='CASCADE'),
        sa.PrimaryKeyConstraint('id'),
        sa.UniqueConstraint('priority', 'router_id', name='unique_prio_rid'))

    op.create_table(
        'bsn_nexthops',
        sa.Column('rule_id', sa.Integer(), nullable=False),
        sa.Column('nexthop', sa.String(length=64), nullable=False),
        sa.ForeignKeyConstraint(['rule_id'], ['bsn_routerrules.id'],
                                ondelete='CASCADE'),
        sa.PrimaryKeyConstraint('rule_id', 'nexthop'))


def downgrade():
    pass



# Copyright 2011 OpenStack Foundation.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import sqlalchemy as sa

from networking_bigswitch.plugins.bigswitch.i18n import _
from networking_bigswitch.plugins.bigswitch.utils import Util
from neutron_lib.db import model_base
from neutron_lib.db import model_query
from neutron_lib.db import utils as db_utils
from neutron_lib import exceptions
from oslo_serialization import jsonutils
from sqlalchemy.dialects.mysql.base import VARCHAR
from sqlalchemy.orm import exc
from sqlalchemy.types import Enum, TIMESTAMP, TypeDecorator


class JSONEncodedDict(TypeDecorator):
    """encoded JSON dict

    Represents an immutable structure as a json-encoded string.
    Usage::
        JSONEncodedDict(255)
    """
    impl = VARCHAR

    def process_bind_param(self, value, dialect):
        if value is not None:
            value = jsonutils.dumps(value)
        return value

    def process_result_value(self, value, dialect):
        if value is not None:
            value = jsonutils.loads(value)
        return value


class ReachabilityTest(model_base.BASEV2,
                       model_base.HasId,
                       model_base.HasProject):
    """Reachability Test

    A table to store user configured reachability tests.
    """
    __tablename__ = 'reachabilitytest'
    name = sa.Column(sa.String(64), nullable=False, unique=True)
    src_tenant_id = sa.Column(sa.String(length=36), nullable=True)
    src_tenant_name = sa.Column(sa.String(255), nullable=False)
    src_segment_id = sa.Column(sa.String(length=36), nullable=True)
    src_segment_name = sa.Column(sa.String(255), nullable=True)
    src_ip = sa.Column(sa.String(16), nullable=False)
    dst_ip = sa.Column(sa.String(16), nullable=False)
    expected_result = sa.Column(Enum("dropped by route",
                                     "dropped by policy",
                                     "not permitted by security groups",
                                     "dropped due to private segment",
                                     "dropped due to loop",
                                     "packet in",
                                     "forwarded",
                                     "dropped",
                                     "unspecified source",
                                     "unsupported",
                                     "invalid input",
                                     "inconsistent status",
                                     "no traffic detected",
                                     name="expected_result"), nullable=False)
    test_time = sa.Column(TIMESTAMP(timezone=True), nullable=True)
    test_result = sa.Column(Enum("pass", "fail", "pending"),
                            default="pending", nullable=False)
    detail = sa.Column(JSONEncodedDict(8192), nullable=True)
    logical_path = sa.Column(JSONEncodedDict(8192), nullable=True)
    run_test = sa.Column(sa.Boolean, nullable=False, default=False)

    def get_connection_source(self, unicode_mode=False):
        source = {}
        if unicode_mode:
            if not self.src_tenant_id:
                raise ReachabilityTestUnicodeTenantIdMissing(
                    test_name=self.name)
            source['tenant'] = self.src_tenant_id
            if not self.src_segment_id:
                raise ReachabilityTestUnicodeSegmentIdMissing(
                    test_name=self.name)
            source['segment'] = self.src_segment_id
        else:
            source['tenant'] = Util.format_resource_name(self.src_tenant_name)
            if not self.src_segment_name:
                raise ReachabilityTestSegmentNameMissing(test_name=self.name)
            source['segment'] = Util.format_resource_name(
                self.src_segment_name)
        source['ip'] = self.src_ip
        return source

    def get_connection_destination(self):
        destination = {}
        destination['ip'] = self.dst_ip
        return destination


class ReachabilityTestNotFound(exceptions.NotFound):
    message = _("Reachability Test %(id)s could not be found")


class ReachabilityTestUnicodeTenantIdMissing(exceptions.NeutronException):
    message = _("Source Tenant ID is required when Unicode is ENABLED. "
                "Please update Test %(test_name)s.")


class ReachabilityTestUnicodeSegmentIdMissing(exceptions.NeutronException):
    message = _("Source Segment ID is required when Unicode is ENABLED. "
                "Please update Test %(test_name)s.")


class ReachabilityTestSegmentNameMissing(exceptions.NeutronException):
    message = _("Source Segment Name is required when Unicode is DISABLED. "
                "Please recreate the network and update Test %(test_name)s.")


class ReachabilityTestDbMixin(object):
    # internal methods
    def _make_reachabilitytest_dict(self, reachabilitytest, fields=None):
        return db_utils.resource_fields({
            'id': reachabilitytest.id,
            'tenant_id': reachabilitytest.tenant_id,
            'name': reachabilitytest.name,
            'src_tenant_id': reachabilitytest.src_tenant_id,
            'src_tenant_name': reachabilitytest.src_tenant_name,
            'src_segment_id': reachabilitytest.src_segment_id,
            'src_segment_name': reachabilitytest.src_segment_name,
            'src_ip': reachabilitytest.src_ip,
            'dst_ip': reachabilitytest.dst_ip,
            'expected_result': reachabilitytest.expected_result,
            'test_time': reachabilitytest.test_time,
            'test_result': reachabilitytest.test_result,
            'detail': reachabilitytest.detail,
            'logical_path': reachabilitytest.logical_path,
            'run_test': reachabilitytest.run_test}, fields)

    def _get_reachabilitytest(self, context, id):
        try:
            reachabilitytest = model_query.get_by_id(
                context, ReachabilityTest, id)
        except exc.NoResultFound:
            raise ReachabilityTestNotFound(id=id)
        return reachabilitytest

    # public CRUD methods for Reachability Test
    def get_reachabilitytests(self, context, filters=None, fields=None,
                              sorts=None, limit=None, marker=None,
                              page_reverse=False):
        reachabilitytests = \
            model_query.get_collection(context, ReachabilityTest,
                                       self._make_reachabilitytest_dict,
                                       filters=filters, fields=fields)
        return reachabilitytests

    def get_reachabilitytest(self, context, id, fields=None):
        reachabilitytest = \
            self._get_reachabilitytest(context, id)
        return self._make_reachabilitytest_dict(reachabilitytest, fields)

    def create_reachabilitytest(self, context, reachabilitytest):
        reachabilitytest_data = reachabilitytest['reachabilitytest']
        with context.session.begin(subtransactions=True):
            reachabilitytest = ReachabilityTest(
                tenant_id=reachabilitytest_data['tenant_id'],
                name=reachabilitytest_data['name'],
                src_tenant_id=reachabilitytest_data.get('src_tenant_id'),
                src_tenant_name=reachabilitytest_data['src_tenant_name'],
                src_segment_id=reachabilitytest_data.get('src_segment_id'),
                src_segment_name=reachabilitytest_data['src_segment_name'],
                src_ip=reachabilitytest_data['src_ip'],
                dst_ip=reachabilitytest_data['dst_ip'],
                expected_result=reachabilitytest_data['expected_result'])
            context.session.add(reachabilitytest)
        return self._make_reachabilitytest_dict(reachabilitytest)

    def create_reachabilitytest_withresult(self, context, reachabilitytest):
        reachabilitytest_data = reachabilitytest['reachabilitytest']
        with context.session.begin(subtransactions=True):
            reachabilitytest = ReachabilityTest(
                tenant_id=reachabilitytest_data['tenant_id'],
                name=reachabilitytest_data['name'],
                src_tenant_id=reachabilitytest_data.get('src_tenant_id'),
                src_tenant_name=reachabilitytest_data['src_tenant_name'],
                src_segment_id=reachabilitytest_data.get('src_segment_id'),
                src_segment_name=reachabilitytest_data['src_segment_name'],
                src_ip=reachabilitytest_data['src_ip'],
                dst_ip=reachabilitytest_data['dst_ip'],
                expected_result=reachabilitytest_data['expected_result'],
                test_result=reachabilitytest_data['test_result'],
                detail=reachabilitytest_data['detail'],
                test_time=reachabilitytest_data['test_time'])
            context.session.add(reachabilitytest)
        return self._make_reachabilitytest_dict(reachabilitytest)

    def update_reachabilitytest(self, context, id, reachabilitytest):
        reachabilitytest_data = reachabilitytest['reachabilitytest']
        with context.session.begin(subtransactions=True):
            reachabilitytest = self._get_reachabilitytest(context, id)
            reachabilitytest.update(reachabilitytest_data)
        return self._make_reachabilitytest_dict(reachabilitytest)

    def delete_reachabilitytest(self, context, id):
        with context.session.begin(subtransactions=True):
            reachabilitytest = self._get_reachabilitytest(context, id)
            context.session.delete(reachabilitytest)


class ReachabilityQuickTest(model_base.BASEV2,
                            model_base.HasId,
                            model_base.HasProject):
    """Reachability Quick Test

    A table to store user configured reachability quick tests.
    """
    __tablename__ = 'reachabilityquicktest'
    name = sa.Column(sa.String(64), nullable=False, unique=True)
    src_tenant_id = sa.Column(sa.String(length=36), nullable=True)
    src_tenant_name = sa.Column(sa.String(255), nullable=False)
    src_segment_id = sa.Column(sa.String(length=36), nullable=True)
    src_segment_name = sa.Column(sa.String(255), nullable=True)
    src_ip = sa.Column(sa.String(16), nullable=False)
    dst_ip = sa.Column(sa.String(16), nullable=False)
    expected_result = sa.Column(Enum("dropped by route",
                                     "dropped by policy",
                                     "not permitted by security groups",
                                     "dropped due to private segment",
                                     "dropped due to loop",
                                     "packet in",
                                     "forwarded",
                                     "dropped",
                                     "unspecified source",
                                     "unsupported",
                                     "invalid input",
                                     "inconsistent status",
                                     "no traffic detected",
                                     name="expected_result"), nullable=False)
    test_time = sa.Column(TIMESTAMP(timezone=True), nullable=True)
    test_result = sa.Column(Enum("pass", "fail", "pending"),
                            default="pending", nullable=False)
    detail = sa.Column(JSONEncodedDict(8192), nullable=True)
    logical_path = sa.Column(JSONEncodedDict(8192), nullable=True)
    run_test = sa.Column(sa.Boolean, nullable=False, default=False)
    save_test = sa.Column(sa.Boolean, nullable=False, default=False)

    def get_connection_source(self, unicode_mode=False):
        source = {}
        if unicode_mode:
            if not self.src_tenant_id:
                raise ReachabilityTestUnicodeTenantIdMissing(
                    test_name=self.name)
            source['tenant'] = self.src_tenant_id
            if not self.src_segment_id:
                raise ReachabilityTestUnicodeSegmentIdMissing(
                    test_name=self.name)
            source['segment'] = self.src_segment_id
        else:
            source['tenant'] = Util.format_resource_name(self.src_tenant_name)
            if not self.src_segment_name:
                raise ReachabilityTestSegmentNameMissing(
                    test_name=self.name)
            source['segment'] = Util.format_resource_name(
                self.src_segment_name)
        source['ip'] = self.src_ip
        return source

    def get_connection_destination(self):
        destination = {}
        destination['ip'] = self.dst_ip
        return destination


class ReachabilityQuickTestDbMixin(object):
    # internal methods
    def _make_reachabilityquicktest_dict(self, reachabilityquicktest,
                                         fields=None):
        return db_utils.resource_fields({
            'id': reachabilityquicktest.id,
            'tenant_id': reachabilityquicktest.tenant_id,
            'name': reachabilityquicktest.name,
            'src_tenant_id': reachabilityquicktest.src_tenant_id,
            'src_tenant_name': reachabilityquicktest.src_tenant_name,
            'src_segment_id': reachabilityquicktest.src_segment_id,
            'src_segment_name': reachabilityquicktest.src_segment_name,
            'src_ip': reachabilityquicktest.src_ip,
            'dst_ip': reachabilityquicktest.dst_ip,
            'expected_result': reachabilityquicktest.expected_result,
            'test_time': reachabilityquicktest.test_time,
            'test_result': reachabilityquicktest.test_result,
            'detail': reachabilityquicktest.detail,
            'logical_path': reachabilityquicktest.logical_path,
            'run_test': reachabilityquicktest.run_test}, fields)

    def _get_reachabilityquicktest(self, context, id):
        try:
            reachabilityquicktest = model_query.get_by_id(
                context, ReachabilityQuickTest, id)
        except exc.NoResultFound:
            raise ReachabilityTestNotFound(id=id)
        return reachabilityquicktest

    # public CRUD methods for Reachability Quick Test
    def get_reachabilityquicktests(self, context, filters=None, fields=None,
                                   sorts=None, limit=None, marker=None,
                                   page_reverse=False):
        reachabilityquicktests = \
            model_query.get_collection(context, ReachabilityQuickTest,
                                       self._make_reachabilityquicktest_dict,
                                       filters=filters, fields=fields)
        return reachabilityquicktests

    def get_reachabilityquicktest(self, context, id, fields=None):
        reachabilityquicktest = \
            self._get_reachabilityquicktest(context, id)
        return self._make_reachabilityquicktest_dict(
            reachabilityquicktest, fields)

    def create_reachabilityquicktest(self, context, reachabilityquicktest):
        quicktest_data = reachabilityquicktest['reachabilityquicktest']
        with context.session.begin(subtransactions=True):
            reachabilityquicktest = ReachabilityQuickTest(
                id=quicktest_data['tenant_id'],
                tenant_id=quicktest_data['tenant_id'],
                name=quicktest_data['name'],
                src_tenant_id=quicktest_data.get('src_tenant_id'),
                src_tenant_name=quicktest_data['src_tenant_name'],
                src_segment_id=quicktest_data.get('src_segment_id'),
                src_segment_name=quicktest_data['src_segment_name'],
                src_ip=quicktest_data['src_ip'],
                dst_ip=quicktest_data['dst_ip'],
                expected_result=quicktest_data['expected_result'])
            context.session.add(reachabilityquicktest)
        return self._make_reachabilityquicktest_dict(reachabilityquicktest)

    def update_reachabilityquicktest(self, context, id, reachabilityquicktest):
        reachabilityquicktest_data = \
            reachabilityquicktest['reachabilityquicktest']
        with context.session.begin(subtransactions=True):
            reachabilityquicktest = self._get_reachabilityquicktest(
                context, id)
            reachabilityquicktest.update(reachabilityquicktest_data)
        return self._make_reachabilityquicktest_dict(reachabilityquicktest)

    def delete_reachabilityquicktest(self, context, id):
        with context.session.begin(subtransactions=True):
            reachabilityquicktest = self._get_reachabilityquicktest(
                context, id)
            context.session.delete(reachabilityquicktest)



# Copyright 2014 Big Switch Networks, Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
import copy
import os

import eventlet

from neutron.extensions import securitygroup as ext_sg
from neutron_lib.api.definitions import portbindings
from neutron_lib.callbacks import events
from neutron_lib.callbacks import registry
from neutron_lib.callbacks import resources
from neutron_lib import context as ctx
from neutron_lib.plugins import directory
from neutron_lib.plugins.ml2 import api
from neutron_lib import rpc as lib_rpc
from oslo_config import cfg
from oslo_log import log
import oslo_messaging
from oslo_utils import excutils
from six.moves import http_client

from networking_bigswitch.plugins.bigswitch import config as pl_config
from networking_bigswitch.plugins.bigswitch import constants as bsn_consts
from networking_bigswitch.plugins.bigswitch.i18n import _LE
from networking_bigswitch.plugins.bigswitch.i18n import _LW
from networking_bigswitch.plugins.bigswitch import plugin
from networking_bigswitch.plugins.bigswitch import servermanager
from networking_bigswitch.plugins.bigswitch.utils import Util

EXTERNAL_PORT_OWNER = 'neutron:external_port'
ROUTER_GATEWAY_PORT_OWNER = 'network:router_gateway'

LOG = log.getLogger(__name__)
add_debug_log = plugin.add_debug_log

# time in seconds to maintain existence of vswitch response
CACHE_VSWITCH_TIME = 60


class BigSwitchMechanismDriver(plugin.NeutronRestProxyV2Base,
                               api.MechanismDriver):

    """Mechanism Driver for Big Switch Networks Controller.

    This driver relays the network create, update, delete
    operations to the Big Switch Controller.
    """
    target = oslo_messaging.Target(version='1.1')

    def initialize(self):
        LOG.debug('Initializing driver')

        # register plugin config opts
        pl_config.register_config()
        self.evpool = eventlet.GreenPool(cfg.CONF.RESTPROXY.thread_pool_size)

        # init network ctrl connections
        self.servers = servermanager.ServerPool()
        self.servers.get_topo_function = self._get_all_data_auto

        # perform one forced topo_sync after 60secs. delayed to let plugin
        # initialization complete
        eventlet.spawn_after(60, self.servers.force_topo_sync,
                             **{'check_ts': True})

        self.segmentation_types = ', '.join(cfg.CONF.ml2.type_drivers)
        # if os-net-config is present, attempt to read physnet bridge_mappings
        # from openvswitch_agent.ini
        self.bridge_mappings = {}
        if os.path.isfile(bsn_consts.RH_NET_CONF_PATH):
            self.bridge_mappings = Util.read_ovs_bridge_mappings()

        self.setup_rpc_callbacks()

        LOG.debug("Initialization done")

    def setup_rpc_callbacks(self):
        # Security group operations are split between callback and RPC
        # notifications mechanism.
        # Callbacks are 'only once' and RPC notifications are 'at least once'.
        # In a HA setup, RPC notifications can be called once per overcloud
        # controller.
        # Callback: SG add, SG delete, SG update, SG rule add
        # RPC notifications: SG add, SG delete, SG rule delete
        # In addition, RPC notifications also used for: tenant add,
        # tenant delete, log all other events as FYI.

        # Register callbacks ONLY IF sync_security_groups is True
        if cfg.CONF.RESTPROXY.sync_security_groups:
            # following way to register call back functions start in kilo
            self._create_sg_f = self.bsn_create_sg_callback
            self._delete_sg_f = self.bsn_delete_sg_callback
            self._update_sg_f = self.bsn_update_sg_callback
            self._create_sg_rule_f = self.bsn_create_sg_rule_callback
            registry.subscribe(self._create_sg_f,
                               resources.SECURITY_GROUP, events.AFTER_CREATE)
            registry.subscribe(self._delete_sg_f,
                               resources.SECURITY_GROUP, events.AFTER_DELETE)
            registry.subscribe(self._update_sg_f,
                               resources.SECURITY_GROUP, events.AFTER_UPDATE)
            registry.subscribe(self._create_sg_rule_f,
                               resources.SECURITY_GROUP_RULE,
                               events.AFTER_CREATE)

        # the above does not cover the cases where security groups are
        # initially created or when they are deleted since those actions
        # aren't needed by the L2 agent. In order to receive those, we
        # subscribe to the notifications topic that receives all of the
        # API create/update/delete events.
        # Notifications are published at the 'info' level so they will result
        # in a call to the 'info' function below. From there we can check
        # the event type and determine what to do from there.
        target = oslo_messaging.Target(topic='#',
                                       server=cfg.CONF.host)
        keystone_target = oslo_messaging.Target(
            topic='#', exchange='keystone', server=cfg.CONF.host)
        self.listener = oslo_messaging.get_notification_listener(
            lib_rpc.TRANSPORT, [target, keystone_target], [self],
            executor='eventlet', allow_requeue=False)
        self.listener.start()

    def bsn_create_sg_callback(self, resource, event, trigger, **kwargs):
        security_group = kwargs.get('security_group')
        context = kwargs.get('context')
        if security_group and context:
            sg_id = security_group.get('id')
            LOG.debug("Callback create sg_id: %s", sg_id)
            self.bsn_create_security_group(sg_id=sg_id, context=context)

    def bsn_delete_sg_callback(self, resource, event, trigger, **kwargs):
        sg_id = kwargs.get('security_group_id')
        context = kwargs.get('context')
        if sg_id and context:
            LOG.debug("Callback delete sg_id: %s", sg_id)
            self.bsn_delete_security_group(sg_id=sg_id, context=context)

    def bsn_update_sg_callback(self, resource, event, trigger, **kwargs):
        security_group = kwargs.get('security_group')
        context = kwargs.get('context')
        if security_group and context:
            sg_id = security_group.get('id')
            LOG.debug("Callback update sg_id: %s", sg_id)
            self.bsn_create_security_group(sg_id=sg_id, context=context)

    def bsn_create_sg_rule_callback(self, resource, event, trigger, **kwargs):
        rule = kwargs.get('security_group_rule')
        context = kwargs.get('context')
        if rule and context:
            sg_id = rule.get('security_group_id')
            LOG.debug("Callback create rule in sg_id: %s", sg_id)
            self.bsn_create_security_group(sg_id=sg_id, context=context)

    def bsn_delete_sg_rule(self, sg_rule, context):
        LOG.debug("Deleting security group rule from CCF: %s", sg_rule)
        if not context:
            LOG.error(_LE(
                "Context missing when trying to delete security group rule. "
                "Please force-ccf-sync to ensure consistency with CCF."))
        sg_id = sg_rule['security_group_id']
        # we over write the sg on bcf controller instead of deleting
        try:
            self.bsn_create_security_group(sg_id,
                                           context=context)
        except ext_sg.SecurityGroupNotFound:
            # DB query will throw exception when security group is
            # being deleted. delete_security_group_rule callback would
            # try to update BCF with new set of rules.
            LOG.warning(
                _LW("Security group with ID %(sg_id)s not found "
                    "when trying to update."), {'sg_id': sg_id})

    def info(self, ctxt, publisher_id, event_type, payload, metadata):
        """This is called on each notification to the neutron topic """
        # we retain this section for security groups, because it handles
        # other events as well. Ignore security group events if disabled in
        # config
        if event_type == 'security_group.create.end':
            LOG.debug("Security group created: %s", payload)
            if cfg.CONF.RESTPROXY.sync_security_groups:
                self.bsn_create_security_group(sg=payload['security_group'])
        elif event_type == 'security_group.delete.end':
            LOG.debug("Security group deleted: %s", payload)
            if cfg.CONF.RESTPROXY.sync_security_groups:
                self.bsn_delete_security_group(payload['security_group_id'])
        elif event_type == 'security_group_rule.delete.end':
            LOG.debug("Security group rule deleted: %s", payload)
            if cfg.CONF.RESTPROXY.sync_security_groups:
                self.bsn_delete_sg_rule(payload['security_group_rule'], ctxt)
        elif event_type == 'identity.project.deleted':
            LOG.debug("Project deleted: %s", payload)
            self.bsn_delete_tenant(payload['resource_info'])
        elif event_type == 'identity.project.created':
            LOG.debug("Project created: %s", payload)
            self.bsn_create_tenant(payload['resource_info'])
        elif event_type == 'identity.project.updated':
            LOG.debug("Project updated: %s", payload)
            # update is the same as create, nsapi will handle it
            self.bsn_create_tenant(payload['resource_info'])
        else:
            LOG.debug("Else events: %s payload: %s", (event_type, payload))

    @add_debug_log
    def security_groups_rule_updated(self, context, **kwargs):
        # this will get called whenever a security group rule updated message
        # goes onto the RPC bus
        LOG.debug("security_groups_rule_updated: %s", kwargs)
        if kwargs.get('security_groups'):
            for sg_id in kwargs.get('security_groups'):
                self.bsn_create_security_group(sg_id, context=context)

    @add_debug_log
    def security_groups_member_updated(self, context, **kwargs):
        pass

    @add_debug_log
    def security_groups_provider_updated(self, context, **kwargs):
        pass

    @add_debug_log
    def create_network_postcommit(self, context):
        # create network on the network controller
        self._send_create_network(context.current)

    @add_debug_log
    def update_network_precommit(self, context):
        self._verify_network_precommit(context)

    @add_debug_log
    def update_network_postcommit(self, context):
        # update network on the network controller
        self._send_update_network(context.current)

    @add_debug_log
    def update_subnet_postcommit(self, context):
        self._trigger_network_update_from_subnet_transaction(context)

    @add_debug_log
    def create_subnet_postcommit(self, context):
        self._trigger_network_update_from_subnet_transaction(context)

    @add_debug_log
    def delete_subnet_postcommit(self, context):
        self._trigger_network_update_from_subnet_transaction(context)

    def _trigger_network_update_from_subnet_transaction(self, context):
        net = directory.get_plugin().get_network(
            ctx.get_admin_context(), context.current['network_id'])
        self._send_update_network(net)

    @add_debug_log
    def delete_network_postcommit(self, context):
        # delete network on the network controller
        self._send_delete_network(context.current)

    @add_debug_log
    def create_port_postcommit(self, context):
        if not self._is_port_supported(context.current):
            LOG.debug("Ignoring unsupported vnic type")
            return

        if self._is_port_sriov(context.current):
            LOG.debug("SR-IOV port, nothing to do")
            return

        try:
            # create port on the network controller
            port = self._prepare_port_for_controller(context)
        except servermanager.TenantIDNotFound as e:
            LOG.warning("Skipping create port %(port)s as %(exp)s",
                        {'port': context.current.get('id'), 'exp': e})
            return

        if port:
            # For vhostuser type ports, membership rule and endpoint was
            # created during bind_port, so skip this
            if port[portbindings.VIF_TYPE] == portbindings.VIF_TYPE_VHOST_USER:
                return

            self.async_port_create(port["network"]["tenant_id"],
                                   port["network"]["id"], port)

    @add_debug_log
    def update_port_postcommit(self, context):
        if not self._is_port_supported(context.current):
            LOG.debug("Ignoring unsupported vnic type")
            return

        # OSP-68: check if port is SRIOV and VM detach case, then skip host_id
        # check and delete port on controller side
        # read-only for shared context okay. deepcopy before modifying
        port = context.current
        network = context.network.current
        if self._is_port_sriov_vm_detach(port, network):
            LOG.debug("update_port_postcommmit called for SRIOV port VM "
                      "detach case.")
            # remove port from BCF and return
            self.servers.rest_delete_port(network["tenant_id"],
                                          network["id"],
                                          port["id"])
            return

        # Else: regular port update,
        # update port on the network controller
        try:
            port = self._prepare_port_for_controller(context)
        except servermanager.TenantIDNotFound as e:
            LOG.warning("Skipping update port %(port)s as %(exp)s",
                        {'port': context.current.get('id'), 'exp': e})
            return

        if port:
            try:
                # For SR-IOV ports, we shouldn't update the port status
                update_status = not self._is_port_sriov(port)
                self.async_port_create(port["network"]["tenant_id"],
                                       port["network"]["id"], port,
                                       update_status)
            except servermanager.RemoteRestError as e:
                with excutils.save_and_reraise_exception() as ctxt:
                    if (cfg.CONF.RESTPROXY.auto_sync_on_failure and
                            e.status == http_client.NOT_FOUND and
                            servermanager.NXNETWORK in e.reason):
                        ctxt.reraise = False
                        LOG.error("Inconsistency with backend controller "
                                  "triggering full synchronization.")
                        self._send_all_data_auto(
                            triggered_by_tenant=port["network"]["tenant_id"]
                        )

    @add_debug_log
    def delete_port_postcommit(self, context):
        if not self._is_port_supported(context.current):
            LOG.debug("Ignoring unsupported vnic type")
            return

        # delete port on the network controller
        port = context.current
        net = context.network.current
        tenant_id = net['tenant_id']
        if not tenant_id:
            tenant_id = servermanager.SERVICE_TENANT
        self.servers.rest_delete_port(tenant_id, net["id"], port['id'])

    def _prepare_port_for_controller(self, context):
        """Make a copy so the context isn't changed for other drivers

        :exception can throw servermanager.TenantIDNotFound
        """
        port = copy.deepcopy(context.current)
        net = context.network.current
        port['network'] = net
        port['bound_segment'] = context.top_bound_segment
        prepped_port = self._map_display_name_or_tenant(port)
        if prepped_port.get('description'):
            del (prepped_port['description'])
        if self.servers.is_unicode_enabled():
            prepped_port['name'] = None
        prepped_port = self._map_state_and_status(prepped_port)
        prepped_port = self._map_port_hostid(prepped_port, net)
        return prepped_port
